import numpy as np
cimport numpy as np

cdef extern from "lodepng.h":
  void lodepng_encode24(unsigned char** out, size_t* outsize, const unsigned char* image, unsigned w, unsigned h)

def encode_png(np.ndarray[np.uint8_t, ndim=3] image):
  # cdef np.ndarray[np.uint8_t, ndim=3] uint8_image
  # cdef unsigned char* uint8_image_data
  # cdef unsigned int height = image.shape[0]
  # cdef unsigned int width = image.shape[1]
  #
  # uint8_image = np.ascontiguousarray(image.ravel(), dtype=np.uint8)
  # uint8_image_data = <unsigned char*>uint8_image.data
  #
  #

  return None


def decode_png(image):
  return None


# -*- coding: UTF-8 -*-
# @Time    : 2019/1/22 1:16 PM
# @File    : activelearning.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function


from antgo.resource.html import *
from antgo.ant.base import *
from antgo.activelearning.samplingmethods.kcenter_greedy import *
from antgo.crowdsource.activelearning_server import *
from antgo.dataflow.common import *
from antgo.dataflow.recorder import *
from antvis.client.httprpc import *
from multiprocessing import Process, Queue
from antgo.task.task import *
from scipy.stats import entropy
import subprocess
import os
import socket
import requests
import json


def _is_open(check_ip, port):
  s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  try:
    s.connect((check_ip, int(port)))
    s.shutdown(2)
    return True
  except:
    return False


def _pick_idle_port(from_port=40000, check_count=100):
  check_port = from_port
  while check_count:
    if not _is_open('127.0.0.1', check_port):
      break

    logger.warn('port %d is occupied, try to use %d port'%(int(check_port), int(check_port + 1)))

    check_port += 1
    check_count -= 1

    if check_count == 0:
      check_port = None

  if check_port is None:
    logger.warn('couldnt find valid free port')
    exit(-1)

  return check_port


class AntActiveLearning(AntBase):
  def __init__(self,
               ant_context,
               ant_name,
               ant_data_folder,
               ant_dump_dir,
               ant_token,
               ant_task_config=None,
               **kwargs):
    super(AntActiveLearning, self).__init__(ant_name, ant_context, ant_token, **kwargs)

    self.skip_first_training = kwargs.get('skip_training', False)
    self.max_time = kwargs.get('max_time', '1d')
    self.max_iterators = getattr(self.context.params, 'max_iterators', 100)
    self.min_label_ratio = getattr(self.context.params, 'min_label_ratio', 0.1)
    self.dump_dir = ant_dump_dir

    self.web_server_port = kwargs.get('port', None)
    self.web_server_port = int(self.web_server_port) if self.web_server_port is not None else None
    self.html_template = kwargs.get('html_template', None)
    self.keywords_template = {}

    self.ant_task_config = ant_task_config
    self.task = kwargs.get('task', '')
    self.ant_data_source = ant_data_folder
    self.option = kwargs.get('option', '')
    self.root_expeirment = kwargs.get('from_experiment', None)
    self.devices = kwargs.get('devices', '')
    self.rpc = None

  def _core_set_algorithm(self, unlabeled_pool, num):
    feature_data = np.array([data['feature'] for data in unlabeled_pool])
    channels = feature_data.shape[-1]
    kcg = kCenterGreedy(feature_data.reshape(-1, channels))
    next_selected = kcg.select_batch(model=None, already_selected=[], N=num)
    next_selected = [unlabeled_pool[int(s)] for s in next_selected]

    return next_selected

  def _entroy_algorithm(self, unlabeled_pool, num):
    unlabeled_entropy = []
    for index, data in enumerate(unlabeled_pool):
      h = np.histogram(data['feature'].flatten(), 255)
      p = h[0].astype(float) / h[0].sum()   # probability of bins
      unlabeled_entropy.append((entropy(p), index))

    ordered_unlabeled = sorted(unlabeled_entropy, key=lambda x: x[0], reverse=True)
    next_selected = [unlabeled_pool[s[1]] for s in ordered_unlabeled[0:num]]
    return next_selected

  def _smart_select(self, unlabeled_pool, num):
    if self.option.lower() == 'coreset':
      return self._core_set_algorithm(unlabeled_pool, num)
    elif self.option.lower() == 'entropy':
      return self._entroy_algorithm(unlabeled_pool, num)
    else:
      return self._entroy_algorithm(unlabeled_pool, num)

  def start(self):
    # 0.step loading challenge task
    running_ant_task = None
    if self.token is not None:
      # 0.step load challenge task
      response = self.context.dashboard.challenge.get(command=type(self).__name__)
      if response['status'] is None:
        # invalid token
        logger.error('couldnt load challenge task')
        self.token = None
      elif response['status'] == 'SUSPEND':
        # prohibit submit challenge task frequently
        # submit only one in one week
        logger.error('prohibit submit challenge task frequently')
        exit(-1)
      elif response['status'] == 'OK':
        # maybe user token or task token
        content = response['content']
        if 'task' in content:
          challenge_task = create_task_from_json(content)
          if challenge_task is None:
            logger.error('couldnt load challenge task')
            exit(-1)
          running_ant_task = challenge_task
      else:
        # unknow error
        logger.error('unknow error')
        exit(-1)

    if running_ant_task is None:
      # 0.step load custom task
      custom_task = create_task_from_xml(self.ant_task_config, self.context)
      if custom_task is None:
        logger.error('couldnt load custom task')
        exit(0)
      running_ant_task = custom_task

    assert (running_ant_task is not None)

    # 配置html模板信息
    self.keywords_template['TASK_TITLE'] = running_ant_task.task_name
    self.keywords_template['TASK_TYPE'] = running_ant_task.task_type
    self.keywords_template['UNLABELED_NUM'] = 0
    self.keywords_template['LABEL_COMPLEX'] = 0

    # dataset
    dataset = running_ant_task.dataset('train',
                                       os.path.join(self.ant_data_source, running_ant_task.dataset_name),
                                       running_ant_task.dataset_params)

    # prepare workspace
    if not os.path.exists(os.path.join(self.main_folder, 'web', 'static', 'data')):
      os.makedirs(os.path.join(self.main_folder, 'web', 'static', 'data'))

    annotation_folder = os.path.join(self.main_folder, 'web', 'static', 'data', 'annotations')
    if not os.path.exists(annotation_folder):
      os.makedirs(annotation_folder)

    data_folder = os.path.join(self.main_folder, 'web', 'static', 'data', 'images')
    if not os.path.exists(data_folder):
      os.makedirs(data_folder)

    download_folder = os.path.join(self.main_folder, 'web', 'static', 'data', 'download')
    if not os.path.exists(download_folder):
      os.makedirs(download_folder)

    upload_folder = os.path.join(self.main_folder, 'web', 'static', 'data', 'upload')
    if not os.path.exists(upload_folder):
      os.makedirs(upload_folder)

    # 数据队列
    request_queue = Queue()

    # launch web server
    if self.web_server_port is None:
      self.web_server_port = 10000
    self.web_server_port = _pick_idle_port(self.web_server_port)

    logger.info('launch active learning web server on port %d'%self.web_server_port)
    process = multiprocessing.Process(target=activelearning_web_server,
                                      args=('activelearning',
                                            self.main_folder,
                                            self.html_template,
                                            self.keywords_template,
                                            running_ant_task,
                                            self.web_server_port,
                                            os.getpid(),
                                            download_folder,
                                            upload_folder,
                                            request_queue))
    process.daemon = True
    process.start()
    logger.info('waiting 5 seconds for launching web server')
    time.sleep(5)

    self.rpc = HttpRpc('v1','activelearning','127.0.0.1',self.web_server_port)
    avg_analyze_time = 0

    # prepare waiting unlabeled data
    try_iter = 0
    while try_iter < self.max_iterators:
      unlabeled_dataset_size = dataset.unlabeled_size()
      labeled_dataset_size = dataset.candidates_size()
      if unlabeled_dataset_size == 0:
        logger.info('active learning is over')
        return

      if not os.path.exists(self.dump_dir):
        os.makedirs(self.dump_dir)

      # 1.step training using labeled data
      for f in os.listdir(self.dump_dir):
        if f[0] == '.':
          continue

        if f == self.root_expeirment:
          continue

        shutil.rmtree(os.path.join(self.dump_dir, f))

      # 通知处理状态(开始未标注数据集挖掘)
      self.rpc.state.patch(round=try_iter,
                           process_state="UNLABEL-PREPARE")
      # 开始分析时间
      analyze_start_time = time.time()
      experiment_id = self.root_expeirment
      if not self.skip_first_training or try_iter > 0:
        # shell call
        logger.info('start training using all labeled data (%d iter)'%try_iter)
        cmd_shell = 'antgo train --main_file=%s --main_param=%s' % (self.main_file, self.main_param)
        cmd_shell += ' --max_time=%s' % self.max_time
        cmd_shell += ' --dump=%s' % self.dump_dir
        cmd_shell += ' --main_folder=%s' % self.main_folder
        cmd_shell += ' --task=%s' % self.task
        if self.root_expeirment is not None:
          cmd_shell += ' --from_experiment=%s' % self.root_expeirment
        cmd_shell += ' --devices=%s' % self.devices
        cmd_shell += ' --name=%s' % self.ant_name
        if self.token is not None:
          cmd_shell += ' --token=%s' % self.token
        training_p = subprocess.Popen('%s > %s.log' % (cmd_shell, self.name), shell=True, cwd=self.main_folder)

        # waiting untile finish training
        training_p.wait()

        # 根据返回结果，判断是否正常结束
        if training_p.returncode != 0:
          logger.error('training process exit anomaly')
          exit(-1)

        # 获取训练完成后的实验目录
        experiment_id = None
        for f in os.listdir(self.dump_dir):
          if f[0] == '.':
            continue

          if f == self.root_expeirment:
            continue

          if os.path.isdir(os.path.join(self.dump_dir, f)):
            experiment_id = f

        if experiment_id is None:
          logger.error('couldnt finding training experiment, maybe training and candidate dataset error')
          exit(-1)

      # 2.step inference using unlabeled data
      logger.info('start analyze all unlabeled data distribution (%d iter)'%try_iter)
      cmd_shell = 'antgo batch --main_file=%s --main_param=%s'%(self.main_file, self.main_param)
      cmd_shell += ' --main_folder=%s' % self.main_folder
      cmd_shell += ' --dump=%s' % self.dump_dir
      if experiment_id is not None:
        cmd_shell += ' --from_experiment=%s' % experiment_id
      cmd_shell += ' --task=%s' % self.task
      cmd_shell += ' --unlabel'
      cmd_shell += ' --devices=%s' % self.devices
      cmd_shell += ' --name=%s' % self.ant_name
      if self.token is not None:
        cmd_shell += ' --token=%s' % self.token
      inference_p = subprocess.Popen('%s > %s.log' %(cmd_shell, self.name), shell=True, cwd=self.main_folder)
      # all processed data are saved in dump/self.name/record/

      # waiting untile finish inference
      inference_p.wait()

      # 根据返回结果，判断是否正常结束
      if inference_p.returncode != 0:
        logger.error('inference process exit anomaly')
        exit(-1)

      # 获取推断完成后的实验目录
      inference_experiment_id = None
      for f in os.listdir(self.dump_dir):
        if f[0] == '.':
          continue

        if os.path.isdir(os.path.join(self.dump_dir, f)):
          if f != experiment_id and f != self.root_expeirment:
            inference_experiment_id = f

      if inference_experiment_id is None:
        logger.error('couldnt finding inference result on unlabeled dataset, maybe unlabeled dataset error')
        exit(-1)

      record_reader = RecordReader(os.path.join(self.dump_dir, inference_experiment_id, 'record'))
      unlabeled_pool = []
      for ss in record_reader.iterate_read('groundtruth', 'predict'):
        gt, feature = ss
        data_file_id = gt['file_id']
        unlabeled_pool.append({'file_id': data_file_id, 'feature': feature})

      select_size = int(len(unlabeled_pool) * self.min_label_ratio)
      if select_size == 0:
        select_size = len(unlabeled_pool)

      next_selected = self._smart_select(unlabeled_pool, select_size)
      round_size = len(next_selected)

      # 结束分析时间
      analyze_end_time = time.time()

      # 获得平均分析时间
      avg_analyze_time = (avg_analyze_time * try_iter + (analyze_end_time - analyze_start_time)) / (try_iter + 1)

      next_unlabeled_sample_ids = []
      for f in next_selected:
        next_unlabeled_sample_ids.append(f['file_id'])

      # tar target folder
      tar_file = "round_%d.tar.gz"%try_iter
      tar_path = os.path.join(download_folder, tar_file)
      if os.path.exists(tar_path):
        os.remove(tar_path)

      tar = tarfile.open(tar_path, "w:gz")
      for next_unlabeled_sample_id in next_unlabeled_sample_ids:
        tar.add(os.path.join(dataset.unlabeled_folder, next_unlabeled_sample_id),
                arcname="round_%d/%s"%(try_iter,next_unlabeled_sample_id))
      tar.close()

      # 通知处理状态(已经准备好未标注数据集)
      self.rpc.state.patch(round=try_iter,
                           process_state="UNLABEL-RESET",
                           unlabel_dataset=tar_file,
                           unlabeled_size=unlabeled_dataset_size,
                           labeled_size=labeled_dataset_size,
                           round_size=round_size)

      while True:
        logger.info('waiting label (human in loop) in round %d'%try_iter)
        request_content = request_queue.get()

        # 1.step 解压数据集文件
        logger.info("untar label dataset")
        request_label_dataset = request_content['FILE']
        request_round = request_content['ROUND']
        if request_round != try_iter:
          logger.error('request label round not consistent')
          self.rpc.state.patch(round=try_iter,
                               process_state="LABEL-ERROR")
          continue

        # 2.step 建立文件夹
        if os.path.exists(os.path.join(upload_folder, "round_%d"%request_round)):
          shutil.rmtree(os.path.join(upload_folder, "round_%d"%request_round))

        # uncompress
        tar = tarfile.open(os.path.join(upload_folder, request_label_dataset))
        tar.extractall(upload_folder)
        tar.close()

        # 3.step 检查数据集标准是否符合标准
        logger.info("check labeled dataset format")
        is_ok = dataset.check_candidate(next_unlabeled_sample_ids, os.path.join(upload_folder, "round_%d"%request_round))
        if not is_ok:
          logger.error("dataset format maybe error, need to update label")

          self.rpc.state.patch(round=try_iter,
                               process_state="LABEL-ERROR")
          continue

        break

      self.rpc.state.patch(round=try_iter,
                           process_state='LABEL-FINISH',
                           next_round_waiting=avg_analyze_time)

      logger.info('finish round %d label, start next round'%try_iter)
      for sample_file in next_unlabeled_sample_ids:
        if os.path.exists(os.path.join(upload_folder, 'round_%d'%try_iter, sample_file)):
          dataset.make_candidate(sample_file,
                                 os.path.join(upload_folder, 'round_%d'%try_iter, sample_file),
                                 'OK')

      # increment round
      try_iter += 1



# -*- coding: UTF-8 -*-
# @Time : 2018/8/13
# @File : multil_task.py
# @Author: Jian <jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.task.task import *
from antgo.measures.base import *
import numpy as np

default = {'AntAccuracyMultiL': ('MULTILABEL_ACCURACY', 'MULTILABEL'),
           'AntHammingLossMultiL': ('MULTILABEL_HAMMINGLOSS', 'MULTILABEL')}

class AntAccuracyMultiL(AntMeasure):
  def __init__(self, task):
    super(AntAccuracyMultiL, self).__init__(task, 'MULTILABEL_ACCURACY')
    assert(task.task_type == 'MULTILABEL')
    self.is_support_rank = True

  def eva(self, data, label):
    '''

    :param data: (list) predict (N x [])
    :param label: (list) ground truth label (N x [])
    :return:
    '''

    if label is not None:
      data = zip(data, label)

    count = 0
    accuracy = 0.0
    precision = 0.0
    recall = 0.0
    f1_measure = 0.0
    sample_scores = []

    for predict, gt in data:
      predict_set = set(np.array(predict).astype(np.uint32).tolist())

      gt_label = gt
      id = None
      if type(gt) == dict:
        gt_label = gt['category_id']
        id = gt['id']

      gt_set = set(np.array(gt_label).astype(np.uint32))

      and_num = len(predict_set & gt_set)
      or_num = len(predict_set | gt_set)

      accuracy += float(and_num) / float(or_num)
      precision += float(and_num) / float(len(predict_set))
      recall += float(and_num) / float(len(gt_set))
      f1_measure += (2.0 * float(and_num)) / (float(len(gt_set)) + float(len(predict_set)))

      if id is not None:
        score_list = [accuracy, precision, recall, f1_measure]
        s = score_list[self.support_rank_index]
        sample_scores.append({'id': id, 'score': s, 'category': gt_label})

      count += 1

    accuracy = accuracy / float(count)
    precision = precision / float(count)
    recall = recall / float(count)
    f1_measure = f1_measure / float(count)

    return {'statistic': {'name': self.name,
                          'value': [{'name': 'Accuracy', 'value': accuracy, 'type': 'SCALAR'},
                                    {'name': 'Precision', 'value': precision, 'type': 'SCALAR'},
                                    {'name': 'Recall', 'value': recall, 'type': 'SCALAR'},
                                    {'name': 'F1', 'value': f1_measure, 'type': 'SCALAR'}]},
            'info': sample_scores}


class AntHammingLossMultiL(AntMeasure):
  def __init__(self, task):
    super(AntHammingLossMultiL, self).__init__(task, 'MULTILABEL_HAMMINGLOSS')
    assert(task.task_type == 'MULTILABEL')
    self.is_support_rank = True

  def eva(self, data, label):
    '''

    :param data: (list) predict (N x [])
    :param label: (list) ground truth label (N x [])
    :return:
    '''

    if label is not None:
      data = zip(data, label)

    count = 0
    score = 0.0
    sample_scores = []

    for predict, gt in data:
      gt_label = gt
      id = None
      if type(gt) == dict:
        gt_label = gt['category_id']
        id = gt['id']

      # transform to onehot
      onehot_predict = np.zeros((len(self.task.class_label)), np.int32)
      onehot_predict[predict] = 1

      onehot_gt = np.zeros(len(self.task.class_label))
      onehot_gt[gt_label] = 1

      num = np.count_nonzero(onehot_predict != onehot_gt)
      score += num

      if id is not None:
        sample_scores.append({'id': id, 'score': score, 'category': gt_label})

      count += 1

    score = score / float(count)
    return {'statistic': {'name': self.name,
                          'value': [{'name': self.name, 'value': score, 'type': 'SCALAR'},]},
            'info': sample_scores}


# -*- coding: UTF-8 -*-
# @Time    : 17-12-14
# @File    : simpleimages.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.dataflow.dataset.dataset import *
import os
import copy

__all__ = ['SimpleImages']


class SimpleImages(Dataset):
  def __init__(self, train_or_test, dir=None, params=None):
    super(SimpleImages, self).__init__(train_or_test, dir, params)

    self.data_files = []
    self.label_files = []
    label_flag = getattr(self, 'flag', None)
    if os.path.exists(os.path.join(self.dir, self.train_or_test, 'data')) and \
        os.path.exists(os.path.join(self.dir, self.train_or_test, 'label')):
      # data , label (only support images)
      for data_file in os.listdir(os.path.join(self.dir, self.train_or_test, 'data')):
        if data_file.lower().split('.')[-1] in ['png', 'jpg', 'jpeg']:
          label_file = data_file if label_flag is None else '%s%s.%s'%(data_file.split('.')[0], label_flag, data_file.split('.')[-1])
          label_file_prefix = '.'.join(label_file.split('.')[0:-1])

          if os.path.exists(os.path.join(self.dir, self.train_or_test, 'label', label_file_prefix+'.png')):
            data_file_path = os.path.join(self.dir, self.train_or_test, 'data', data_file)
            label_file_path = os.path.join(self.dir, self.train_or_test, 'label', label_file_prefix+'.png')
            self.data_files.append(data_file_path)
            self.label_files.append(label_file_path)
          elif os.path.exists(os.path.join(self.dir, self.train_or_test, 'label', label_file_prefix+'.jpg')):
            data_file_path = os.path.join(self.dir, self.train_or_test, 'data', data_file)
            label_file_path = os.path.join(self.dir, self.train_or_test, 'label', label_file_prefix+'.jpg')
            self.data_files.append(data_file_path)
            self.label_files.append(label_file_path)
          elif os.path.exists(os.path.join(self.dir, self.train_or_test, 'label', label_file_prefix+'.jpeg')):
            data_file_path = os.path.join(self.dir, self.train_or_test, 'data', data_file)
            label_file_path = os.path.join(self.dir, self.train_or_test, 'label', label_file_prefix+'.jpeg')
            self.data_files.append(data_file_path)
            self.label_files.append(label_file_path)
    else:
      # data
      for data_file in os.listdir(os.path.join(self.dir, self.train_or_test)):
        if data_file.lower().split('.')[-1] in ['png', 'jpg', 'jpeg']:
          file_path = os.path.join(self.dir, self.train_or_test, data_file)
          self.data_files.append(file_path)
    
    assert(len(self.data_files) > 0)
    self.ids = [i for i in range(len(self.data_files))]
  
  def data_pool(self):
    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1

      ids = copy.copy(self.ids)
      if self.rng:
        self.rng.shuffle(ids)
      
      for id in ids:
        img = imread(self.data_files[id])
        if len(self.label_files) > 0:
          label = imread(self.label_files[id])
          seg_map = label
          if len(label.shape) == 3:
            seg_map = label[:,:,0]
          annotation = {'segmentation_map': seg_map, 'data':label, 'id': id, 'path': self.data_files[id]}
          yield [img, annotation]
        else:
          annotation = {'id': id, 'path': self.data_files[id]}
          yield [img, annotation]
        
  def at(self, id):
    img = imread(self.data_files[id])
    if len(self.label_files) > 0:
      label = imread(self.label_files[id])
      seg_map = label
      if len(label.shape) == 3:
        seg_map = label[:, :, 0]
      annotation = {'segmentation_map': seg_map, 'data':label, 'id': id, 'path': self.data_files[id]}
      return [img, annotation]
    else:
      annotation = {'id': id, 'path': self.data_files[id]}
      return [img, annotation]

  @property
  def size(self):
    return len(self.ids)

  def split(self, split_params={}, split_method='holdout'):
    assert (self.train_or_test == 'train')
    assert (split_method == 'holdout')

    return self, SimpleImages('val', self.dir, self.ext_params)



# encoding=utf-8
# @Time    : 17-7-31
# @File    : basic.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.utils.serialize import loads, dumps
import numpy as np
import os
import traceback
import sys
import yaml
from antgo import config
import antgo.utils.pickledb as pickledb
from contextlib import contextmanager
from antgo.utils import logger


class Sample(object):
  def __init__(self, **kwargs):
    self.data = kwargs

  def serialize(self):
    return self.data

  @staticmethod
  def unserialize(bytes_data):
    return loads(bytes_data)


class RecordWriter(object):
  def __init__(self, record_path):
    self._record_path = os.path.join(record_path, 'data.db')
    
    try:
      self._db = pickledb.load(self._record_path, False)
      count = self._db.get(str('attrib-count'))
      if count is None:
        self._db.set(str('attrib-count'), str(0))
    except:
      logger.error('couldnt open db')
    
  def close(self):
    self._db.dump()

  def write(self, sample, sample_index=-1):
    count = self._db.get(str('attrib-count'))
    assert(count is not None)
    count = int(count)

    sample_key = str(count)
    if sample_index > 0:
      sample_key = str(sample_index)
    self._db.set(sample_key, sample.serialize())

    count += 1
    self._db.set('attrib-count', str(count))
    
  def bind_attrs(self, **kwargs):
    # bind extra db attributes
    for k,v in kwargs.items():
      self._db.set(str('attrib-%s'%k), str('attrib-%s'%v))
  
  @property
  def size(self):
    count = self._db.get(str('attrib-count'))
    if count is None:
      return 0
    
    count = int(count)
    return count

@contextmanager
def safe_recorder_manager(recorder):
  try:
    yield recorder
    recorder.close()
  except:
    traceback.print_exc()
    raise sys.exc_info()[0]

@contextmanager
def safe_manager(handler):
  try:
    yield handler
  except:
    traceback.print_exc()
    raise sys.exc_info()[0]


class RecordReader(object):
  def __init__(self, record_path, read_only=True):
    # db
    if os.path.exists(os.path.join(record_path, 'record')):
      record_path = os.path.join(record_path, 'record')

    self._record_path = os.path.join(record_path, 'data.db')

    try:
      self._db = pickledb.load(self._record_path, False)
      # db attributes
      self._db_attrs = {}
      self.count = self._db.get('attrib-count')

      # attrib
      # it = self._db.iteritems()
      # it.seek('attrib-')
      # for attrib_item in it:
      #   key, value = attrib_item
      #   if key.startswith('attrib-'):
      #     key = key.replace('attrib-', '')
      #     value = value.replace('attrib-', '')
      #     self._db_attrs[key] = value
      #     setattr(self, key, value)
    except:
      logger.error('couldnt open rocksdb')

  def close(self):
    pass

  def record_attrs(self):
    return self._db_attrs

  def bind_attrs(self, **kwargs):
    # bind extra db attributes
    for k, v in kwargs.items():
      self._db.set(str('attrib-%s' % k), str('attrib-%s' % v))

  def put(self, key, data):
    self._db.set(str(key), str(data))

  def get(self, key):
    return self._db.get(str(key))

  def write(self, sample, sample_index=-1):
    count = self._db.get(str('attrib-count'))
    if count is None:
      self._db.set(str('attrib-count'), '0')
      count = 0

    count = int(count)
    sample_key = str(count)
    if sample_index >= 0 :
      sample_key = str(sample_index)
    self._db.set(sample_key, sample.serialize())

    count += 1
    self._db.set('attrib-count', str(count))

  def read(self, index, *args):
    try:
      data = self._db.get(str(index))
      # data = Sample.unserialize(ss)
      sample = []
      if len(args) == 0:
        for k, v in data.items():
          sample.append(v)
      else:
        for k in args:
          if k in data:
            sample.append(data[k])
          else:
            sample.append(None)

      return sample
    except:
      return [None for _ in range(len(args))]

  def iterate_read(self, *args):
    for k in range(int(self.count)):
      data = self._db.get(str(k))
      # data = Sample.unserialize(ss)
      sample = []
      if len(args) == 0:
        for data_key, data_val in data.items():
          sample.append(data_val)
        yield sample
      else:
        for data_key in args:
          if data_key in data:
            sample.append(data[data_key])
          else:
            sample.append(None)
        yield sample

  def iterate_sampling_read(self, index, *args):
    for i in index:
      data = self._db.get(str(i))
      # data = Sample.unserialize(ss)
      sample = []
      if len(args) == 0:
        for data_key, data_val in data.items():
          sample.append(data_val)
        yield sample
      else:
        for data_key in args:
          if data_key in data:
            sample.append(data[data_key])
          else:
            sample.append(None)
        yield sample

  @property
  def size(self):
    count = self._db.get(str('attrib-count'))
    if count is None:
      return 0

    count = int(count)
    return count




# encoding=utf-8
# @Time    : 17-5-24
# @File    : encode.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import numpy as np
from PIL import Image
try:
    # python 2
    from StringIO import StringIO as BufferIO
except ImportError:
    # python 3
    from io import BytesIO as BufferIO


def png_encode(data, thumbnail=False):
  assert(len(data.shape) == 2 or len(data.shape) == 3)
  if data.dtype != np.uint8:
    # normalized to 255
    max_val = float(np.max(data[:]))
    min_val = float(np.min(data[:]))
    norm_data = (data - min_val) / (max_val - min_val) * 255.0
    data = norm_data.astype(np.uint8)

  # 1.step convert to Image
  image = Image.fromarray(data)

  if thumbnail:
    # 2.step (option) thumbnail
    width, height = image.size
    small_size = width if width < height else height
    ratio = small_size / 64
    small_width = int(width / ratio)
    small_height = int(height / ratio)
    image.thumbnail((small_width, small_height), Image.ANTIALIAS)

  # 2.step retarget to bufferio
  temp = BufferIO()
  image.save(temp, qtables='web_low', format='png')
  content = temp.getvalue()
  temp.close()

  return content


# -*- coding: UTF-8 -*-
# @Time    : 2020-06-25 23:30
# @File    : browser_server.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import tornado.httpserver
import tornado.ioloop
import tornado.options
from tornado.options import define, options
from tornado import web, gen
from tornado import httpclient
import tornado.web
import os
import shutil
import signal
from antgo.crowdsource.base_server import *
from antgo.utils import logger
import traceback
import sys
import uuid
import json
import time


class EntryApiHandler(BaseHandler):
  @gen.coroutine
  def get(self, *args, **kwargs):
    # notify backend and waiting response
    session_id = self.get_cookie('sessionid')
    if session_id is None:
      session_id = str(uuid.uuid4())
      self.set_secure_cookie('sessionid', session_id)

    if session_id not in self.db['user']:
      self.db['user'][session_id] = []

    # browser/screening
    if self.settings['browser_mode'] == 'browser':
      # 浏览模式
      if len(self.db['user'][session_id]) == 0:
        if len(self.db['data']) == 0:
          if self.response_queue.empty():
            # 无新数据，直接返回
            self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
            return

          # 从队列取出，加入数据库
          self.db['data'].append({
            'value': self.response_queue.get(),
            'status': False,
            'time': time.time()
          })

        self.db['user'][session_id].append(0)

      # 获得当前用户浏览位置
      entry_id = self.db['user'][session_id][-1]

      # 构建返回数据
      response_content = {
        'value': self.db['data'][entry_id]['value'],
        'step': len(self.db['user'][session_id]) - 1,
        'tags': self.settings.get('tags', []),
        'operator': [],
        'dataset_flag': self.db['state'],
        'samples_num': self.db['dataset'][self.db['state']]['samples_num'],
        'samples_num_checked': self.db['dataset'][self.db['state']]['samples_num_checked'],
      }

      self.response(RESPONSE_STATUS_CODE.SUCCESS, content=response_content)
      return

    # 审查模式
    if len(self.db['user'][session_id]) == 0:
      if self.response_queue.empty():
        # 无新数据，直接返回
        self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
        return

      # 新用户，从队列中获取新数据并加入用户队列中
      self.db['data'].append({
        'value': self.response_queue.get(),
        'status': False,
        'time': time.time()
      })

      latest_id = len(self.db['data']) - 1
      self.db['user'][session_id].append(latest_id)

    # 获得当前用户待审查数据ID
    entry_id = self.db['user'][session_id][-1]

    # 构建返回数据
    response_content = {
      'value': self.db['data'][entry_id]['value'],
      'step': len(self.db['user'][session_id]) - 1,
      'tags': self.settings.get('tags', []),
      'operator': [],
      'dataset_flag': self.db['state'],
      'samples_num': self.db['dataset'][self.db['state']]['samples_num'],
      'samples_num_checked': self.db['dataset'][self.db['state']]['samples_num_checked'],
    }

    self.response(RESPONSE_STATUS_CODE.SUCCESS, content=response_content)


class PrevApiHandler(BaseHandler):
  @gen.coroutine
  def post(self):
    step = self.get_argument('step', None)
    data = self.get_argument("data", None)
    if step is None or data is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'request parameter wrong')
      return

    session_id = self.get_cookie('sessionid')
    if session_id is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'cookie invalid')
      return

    if session_id not in self.db['user']:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'cookie invalid')
      return

    step = int(step)
    if step < 0 or step >= len(self.db['user'][session_id]):
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'request parameter wrong')
      return

    # 保存当前修改到内存
    data = json.loads(data)
    entry_id = self.db['user'][session_id][step]
    self.db['data'][entry_id] = {'value': data, 'status': True, 'time': time.time()}

    # 保存当前修改到文件
    # 使用 offset
    state = self.db['state']
    offset = self.db['dataset'][state]['offset']
    if not os.path.exists(os.path.join(self.dump_folder, state)):
      os.makedirs(os.path.join(self.dump_folder, state))

    try:
      # 发现数据id
      data_id = None
      for item in data:
        if item['title'] == 'ID':
          data_id = str(item['data'])
      if data_id is None:
        data_id = str(entry_id+offset)

      with open(os.path.join(self.dump_folder, state, '%s.json'%data_id), "w") as file_obj:
        json.dump(data, file_obj)
    except Exception as e:
      print('str(Exception):\t', str(Exception))
      print('str(e):\t\t', str(e))
      print('repr(e):\t', repr(e))
      print('e.message:\t', e.message)

    # 获得当前用户上一步数据
    if step == 0:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'dont have pre step')
      return

    pre_step = step - 1
    pre_entry_id = self.db['user'][session_id][pre_step]
    response_content = {
      'value': self.db['data'][pre_entry_id]['value'],
      'step': pre_step,
      'tags': self.settings.get('tags', []),
      'operator': [],
      'state': self.db['state'],
      'samples_num': self.db['dataset'][self.db['state']]['samples_num'],
      'samples_num_checked': self.db['dataset'][self.db['state']]['samples_num_checked'],
    }
    self.response(RESPONSE_STATUS_CODE.SUCCESS, content=response_content)


class NextApiHandler(BaseHandler):
  @gen.coroutine
  def post(self):
    step = self.get_argument("step", None)
    data = self.get_argument("data", None)
    if step is None or data is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'request parameter wrong')
      return

    session_id = self.get_cookie('sessionid')
    if session_id is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'cookie invalid')
      return

    if session_id not in self.db['user']:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'cookie invalid')
      return

    step = int(step)
    if step < 0 or step >= len(self.db['user'][session_id]):
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'request parameter wrong')
      return

    # 保存当前修改到内存
    data = json.loads(data)
    entry_id = self.db['user'][session_id][step]
    if not self.db['data'][entry_id]['status']:
      self.db['dataset'][self.db['state']]['samples_num_checked'] += 1

    self.db['data'][entry_id] = {'value': data, 'status': True, 'time': time.time()}

    # 保存当前修改到文件
    # 使用 offset
    state = self.db['state']
    offset = self.db['dataset'][state]['offset']
    if not os.path.exists(os.path.join(self.dump_folder, state)):
      os.makedirs(os.path.join(self.dump_folder, state))

    try:
      # 发现数据id
      data_id = None
      for item in data:
        if item['title'] == 'ID':
          data_id = str(item['data'])
      if data_id is None:
        data_id = str(entry_id+offset)

      with open(os.path.join(self.dump_folder, state, '%s.json'%data_id), "w") as file_obj:
        json.dump(data, file_obj)
    except Exception as e:
      print('str(Exception):\t', str(Exception))
      print('str(e):\t\t', str(e))
      print('repr(e):\t', repr(e))
      print('e.message:\t', e.message)

    # 获得用户下一步数据
    if step < len(self.db['user'][session_id]) - 1:
      next_step = step + 1
      next_entry_id = self.db['user'][session_id][next_step]
      response_content = {
        'value': self.db['data'][next_entry_id]['value'],
        'step': next_step,
        'tags': self.settings.get('tags', []),
        'operator': [],
        'state': self.db['state'],
        'samples_num': self.db['dataset'][self.db['state']]['samples_num'],
        'samples_num_checked': self.db['dataset'][self.db['state']]['samples_num_checked'],
      }
      self.response(RESPONSE_STATUS_CODE.SUCCESS, content=response_content)
      return

    if self.settings['browser_mode'] == 'browser':
      # 浏览模式
      if len(self.db['data']) == len(self.db['user'][session_id]):
        if self.response_queue.empty():
          # 无新数据，直接返回
          self.response(RESPONSE_STATUS_CODE.SUCCESS, content={
            'value': self.db['data'][self.db['user'][session_id][-1]]['value'],
            'step': len(self.db['user'][session_id]) - 1,
            'tags': self.settings.get('tags', []),
            'operator': [],
            'state': self.db['state'],
            'samples_num': self.db['dataset'][self.db['state']]['samples_num'],
            'samples_num_checked': self.db['dataset'][self.db['state']]['samples_num_checked'],
          })
          return

        # 加入新数据
        self.db['data'].append({'value': self.response_queue.get(),
                                'status': False,
                                'time': time.time()})

      # 为当前用户分配下一个浏览数据
      next_entry_id = len(self.db['user'][session_id])
      self.db['user'][session_id].append(next_entry_id)

      self.response(RESPONSE_STATUS_CODE.SUCCESS, content={
        'value': self.db['data'][next_entry_id]['value'],
        'step': len(self.db['user'][session_id]) - 1,
        'tags': self.settings.get('tags', []),
        'operator': [],
        'state': self.db['state'],
        'samples_num': self.db['dataset'][self.db['state']]['samples_num'],
        'samples_num_checked': self.db['dataset'][self.db['state']]['samples_num_checked'],
      })

      return

    # 审查模式
    # 当前已经到达用户最后一张图，下一张图必须从队列中获取
    # 新数据来源（1，之前分配给某个用户，但一直没有收到反馈；2，新产生的数据）
    # 1. 之前分配给某个用户，但一直没有收到反馈
    next_entry_id = -1
    now_time = time.time()
    for sample_i, sample in enumerate(self.db['data']):
      if not sample['status'] and (now_time - sample['time']) > 60:
        next_entry_id = sample_i
        break

    # 2. 新产生的数据
    if self.response_queue.empty():
      # 无新数据，直接返回
      self.response(RESPONSE_STATUS_CODE.SUCCESS, content={
        'value': self.db['data'][self.db['user'][session_id][-1]]['value'],
        'step': len(self.db['user'][session_id]) - 1,
        'tags': self.settings.get('tags', []),
        'operator': [],
        'state': self.db['state'],
        'samples_num': self.db['dataset'][self.db['state']]['samples_num'],
        'samples_num_checked': self.db['dataset'][self.db['state']]['samples_num_checked'],
      })
      return

    if next_entry_id == -1:
      self.db['data'].append({'value': self.response_queue.get(), 'status': False, 'time': time.time()})
      next_entry_id = len(self.db['data']) - 1

    # 为当前用户分配下一个审查数据
    self.db['user'][session_id].append(next_entry_id)

    #
    response_content = {
      'value': self.db['data'][next_entry_id]['value'],
      'step': len(self.db['user'][session_id]) - 1,
      'tags': self.settings.get('tags', []),
      'operator': [],
      'state': self.db['state'],
      'samples_num': self.db['dataset'][self.db['state']]['samples_num'],
      'samples_num_checked': self.db['dataset'][self.db['state']]['samples_num_checked'],
    }

    #
    self.response(RESPONSE_STATUS_CODE.SUCCESS, content=response_content)


class FileApiHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    # 获得代码目录结构
    root_path = self.data_folder
    files_tree = [{
      'name': ".",
      'type': "folder",
      'size': "",
      'folder': [],
      'path': ''
    }]
    queue = [files_tree[-1]]
    while len(queue) != 0:
      folder = queue.pop(-1)
      folder_path = os.path.join(root_path, folder['path'])
      for f in os.listdir(folder_path):
        if os.path.isdir(os.path.join(folder_path, f)):
          folder['folder'].append({
            'name': f,
            'type': "folder",
            'size': "",
            'folder': [],
            'path': '%s/%s' % (folder['path'], f) if folder['path'] != '' else f,
          })

          queue.append(folder['folder'][-1])
        else:
          fsize = os.path.getsize(os.path.join(folder_path, f))
          fsize = fsize / 1024.0  # KB
          folder['folder'].append({
            'name': f,
            'type': "file",
            'size': "%0.2f KB" % round(fsize, 2),
            'path': '%s/%s' % (folder['path'], f) if folder['path'] != '' else f,
          })

    response = {
      'files_tree': files_tree
    }
    self.response(RESPONSE_STATUS_CODE.SUCCESS, content=response)


class DownloadApiHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    download_file = self.get_argument('path', None)
    if download_file is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'not set download file')
      return

    download_path = os.path.join(self.data_folder, download_file)
    if not os.path.exists(download_path):
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'not download file')
      return

    # Content-Type
    self.set_header('Content-Type', 'application/octet-stream')
    self.set_header('Content-Disposition', 'attachment; filename=' + download_file)

    buffer_size = 64 * 1024
    with open(download_path, 'rb') as fp:
      content = fp.read()
      content_size = len(content)
      buffer_segments = content_size // buffer_size
      for buffer_seg_i in range(buffer_segments):
        buffer_data = content[buffer_seg_i * buffer_size: (buffer_seg_i + 1) * buffer_size]
        yield self.write(buffer_data)

      yield self.write(content[buffer_segments * buffer_size:])

    self.finish()


class OperatorApiHandler(BaseHandler):
  def post(self):
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class ConfigApiHandler(BaseHandler):
  def post(self):
    offset_config = self.get_argument('offset_config', None)
    if offset_config is not None:
      offset_config = json.loads(offset_config)
      dataset_flag = offset_config['dataset_flag']
      dataset_offset = offset_config['dataset_offset']
      if 'dataset' not in self.db:
        self.db['dataset'] = {}
      if dataset_flag not in self.db['dataset']:
        self.db['dataset'][dataset_flag] = {}

      self.db['dataset'][dataset_flag]['offset'] = dataset_offset

    profile_config = self.get_argument('profile_config', None)
    if profile_config is not None:
      profile_config = json.loads(profile_config)
      dataset_flag = profile_config['dataset_flag']
      samples_num = profile_config['samples_num']
      samples_num_checked = profile_config['samples_num_checked']

      if 'dataset' not in self.db:
        self.db['dataset'] = {}
      if dataset_flag not in self.db['dataset']:
        self.db['dataset'][dataset_flag] = {}

      self.db['dataset'][dataset_flag]['samples_num'] = samples_num
      self.db['dataset'][dataset_flag]['samples_num_checked'] = samples_num_checked

    state = self.get_argument('state', None)
    if state is not None:
      # train,val or test
      self.db['state'] = state


class GracefulExitException(Exception):
  @staticmethod
  def sigterm_handler(signum, frame):
    raise GracefulExitException()


def browser_server_start(data_path,
                         browser_dump_dir,
                         response_queue,
                         tags,
                         server_port,
                         offset_configs,
                         profile_config,
                         browser_mode='screening'):
  # register sig
  signal.signal(signal.SIGTERM, GracefulExitException.sigterm_handler)

  # 0.step define http server port
  define('port', default=server_port, help='run on port')

  try:
    # 静态资源目录
    browser_static_dir = os.path.join(browser_dump_dir, 'browser')
    # 数据数据目录
    browser_dump_dir = os.path.join(browser_dump_dir, 'record')
    if not os.path.exists(browser_dump_dir):
      os.makedirs(browser_dump_dir)

    # 2.step launch web server
    db = {'data': [], 'user': {}, 'dataset': {}}
    for offset_config in offset_configs:
      db['dataset'][offset_config['dataset_flag']] = {'offset':offset_config['dataset_offset']}

    db['dataset'][profile_config['dataset_flag']]['samples_num'] = profile_config['samples_num']
    db['dataset'][profile_config['dataset_flag']]['samples_num_checked'] = profile_config['samples_num_checked']
    db['state'] = profile_config['dataset_flag']

    settings = {
      'static_path': os.path.join(browser_static_dir, 'static'),
      'dump_path': browser_dump_dir,
      'port': server_port,
      'response_queue': response_queue,
      'tags': tags,
      'data_folder': data_path,
      'cookie_secret': str(uuid.uuid4()),
      'db': db,
      'browser_mode': browser_mode
    }

    app = tornado.web.Application(handlers=[(r"/browser-api/prev/", PrevApiHandler),
                                            (r"/browser-api/next/", NextApiHandler),
                                            (r"/browser-api/operator/", OperatorApiHandler),
                                            (r"/browser-api/entry/", EntryApiHandler),
                                            (r"/browser-api/file/", FileApiHandler),
                                            (r"/browser-api/download/", DownloadApiHandler),
                                            (r"/browser-api/config/", ConfigApiHandler),
                                            (r'/(.*)', tornado.web.StaticFileHandler,
                                             {"path": browser_static_dir, "default_filename": "index.html"}),],
                                  **settings)
    http_server = tornado.httpserver.HTTPServer(app)
    http_server.listen(options.port)

    logger.info('demo is providing server on port %d' % server_port)
    tornado.ioloop.IOLoop.instance().start()
    logger.info('demo stop server')
  except GracefulExitException:
    logger.info('demo server exit')
    sys.exit(0)
  except KeyboardInterrupt:
    pass

# browser_server_start()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: fs.py
# Author: jian(jian@mltalker.com)
from __future__ import unicode_literals

import os, sys
from six.moves import urllib
from . import logger

def mkdir_p(dirname):
  """ make a dir recursively, but do nothing if the dir exists"""
  assert dirname is not None
  if dirname == '' or os.path.isdir(dirname):
    return
  try:
    os.makedirs(dirname)
  except OSError as e:
    if e.errno != 17:
      raise e

def download(url, dir, fname=None):
  mkdir_p(dir)
  if fname is None:
    fname = url.split('/')[-1]
  fpath = os.path.join(dir, fname)

  def _progress(count, block_size, total_size):
    sys.stdout.write('\r>> Downloading %s %.1f%%' %
                     (fname,
                         min(float(count * block_size)/ total_size,
                             1.0) * 100.0))
    sys.stdout.flush()
  try:
    fpath, _ = urllib.request.urlretrieve(url, fpath, reporthook=_progress)
    statinfo = os.stat(fpath)
    size = statinfo.st_size
  except:
    logger.error("Failed to download {}".format(url))
    raise
  assert size > 0, "Download an empty file!"
  sys.stdout.write('\n')
  print('Succesfully downloaded ' + fname + " " + str(size) + ' bytes.')
  return fpath

def maybe_here(dest_dir, target_file):
  maybe_dest_dir = dest_dir
  while maybe_dest_dir is not None:
    target_file_path = os.path.join(maybe_dest_dir, target_file)
    if not os.path.exists(target_file_path):
      is_over = False
      for ff in os.listdir(maybe_dest_dir):
        if ff[0] == '.':
          continue
        if os.path.isdir(os.path.join(maybe_dest_dir, ff)):
          maybe_dest_dir = os.path.join(maybe_dest_dir, ff)
          break

        is_over = True

      if is_over:
        maybe_dest_dir = None
    else:
      break

  return maybe_dest_dir

def maybe_here_match_format(dest_dir, target_pattern):
  maybe_dest_dir = dest_dir
  while maybe_dest_dir is not None:
    files = os.listdir(maybe_dest_dir)
    is_has_files = len([f for f in files if target_pattern == f])
    if is_has_files == 0:
      is_continue = False
      for ff in os.listdir(maybe_dest_dir):
        if ff[0] == '.':
          continue
        if os.path.isdir(os.path.join(maybe_dest_dir, ff)):
          maybe_dest_dir = os.path.join(maybe_dest_dir, ff)
          is_continue = True
          break

      if not is_continue:
        maybe_dest_dir = None
      else:
        break
    else:
      break

  return maybe_dest_dir

def maybe_here_fixed_3_hierarchy(dest_dir,target_pattern):
  if not os.path.exists(dest_dir):
    return None

  maybe_dest_dir = dest_dir
  while maybe_dest_dir is not None:
    check_dirs = [os.path.join(maybe_dest_dir, family, character) \
                              for family in os.listdir(maybe_dest_dir) \
                              if os.path.isdir(os.path.join(maybe_dest_dir, family)) \
                              for character in os.listdir(os.path.join(maybe_dest_dir, family))]

    if len(check_dirs) > 0:
      files = os.listdir(check_dirs[0])
      is_has_files = len([f for f in files if target_pattern in f])
      if is_has_files > 0:
        return maybe_dest_dir

    is_continue = False
    for ff in os.listdir(maybe_dest_dir):
      if ff[0] == '.':
        continue
      if os.path.isdir(os.path.join(maybe_dest_dir, ff)):
        maybe_dest_dir = os.path.join(maybe_dest_dir, ff)
        is_continue = True
        break

    if not is_continue:
      maybe_dest_dir = None
      return maybe_dest_dir

if __name__ == '__main__':
    download('http://dl.caffe.berkeleyvision.org/caffe_ilsvrc12.tar.gz', '.')



# -*- coding: UTF-8 -*-
# File: utils.py
# Author: jian <jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import os, sys
from contextlib import contextmanager
import inspect, functools
from datetime import datetime
import time
import collections
import numpy as np
import six
import tarfile
from datetime import datetime
if sys.version > '3':
    PY3 = True
else:
    PY3 = False


__all__ = ['change_env',
        'map_arg',
        'get_rng', 'memoized',
        'get_tqdm_kwargs',
        'targz',
        'timestamp']

#def expand_dim_if_necessary(var, dp):
#    """
#    Args:
#        var: a tensor
#        dp: a numpy array
#    Return a reshaped version of dp, if that makes it match the valid dimension of var
#    """
#    shape = var.get_shape().as_list()
#    valid_shape = [k for k in shape if k]
#    if dp.shape == tuple(valid_shape):
#        new_shape = [k if k else 1 for k in shape]
#        dp = dp.reshape(new_shape)
#    return dp

@contextmanager
def change_env(name, val):
  oldval = os.environ.get(name, None)
  os.environ[name] = val
  yield
  if oldval is None:
    del os.environ[name]
  else:
    os.environ[name] = oldval

class memoized(object):
  '''Decorator. Caches a function's return value each time it is called.
  If called later with the same arguments, the cached value is returned
  (not reevaluated).
  '''
  def __init__(self, func):
    self.func = func
    self.cache = {}

  def __call__(self, *args):
    if not isinstance(args, collections.Hashable):
      # uncacheable. a list, for instance.
      # better to not cache than blow up.
      return self.func(*args)
    if args in self.cache:
      return self.cache[args]
    else:
      value = self.func(*args)
      self.cache[args] = value
      return value

  def __repr__(self):
    '''Return the function's docstring.'''
    return self.func.__doc__

  def __get__(self, obj, objtype):
    '''Support instance methods.'''
    return functools.partial(self.__call__, obj)


#_GLOBAL_MEMOIZED_CACHE = dict()
#def global_memoized(func):
    #""" Make sure that the same `memoized` object is returned on different
        #calls to global_memoized(func)
    #"""
    #ret = _GLOBAL_MEMOIZED_CACHE.get(func, None)
    #if ret is None:
        #ret = _GLOBAL_MEMOIZED_CACHE[func] = memoized(func)
    #return ret

def map_arg(**maps):
  """
  Apply a mapping on certains argument before calling original function.
  maps: {key: map_func}
  """
  def deco(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
      argmap = inspect.getcallargs(func, *args, **kwargs)
      for k, map_func in six.iteritems(maps):
        if k in argmap:
          argmap[k] = map_func(argmap[k])
      return func(**argmap)
    return wrapper
  return deco

def get_rng(obj=None, seed=None):
  """ obj: some object to use to generate random seed"""
  if seed is None:
    seed = (id(obj) + os.getpid() +
            int(datetime.now().strftime("%Y%m%d%H%M%S%f"))) % 4294967295
  return np.random.RandomState(seed)


def get_tqdm_kwargs(**kwargs):
  default = dict(
          smoothing=0.5,
          dynamic_ncols=True,
          ascii=True,
          bar_format='{l_bar}{bar}|{n_fmt}/{total_fmt}[{elapsed}<{remaining},{rate_noinv_fmt}]'
          )
  f = kwargs.get('file', sys.stderr)
  if f.isatty():
    default['mininterval'] = 0.5
  else:
    default['mininterval'] = 60
  default.update(kwargs)
  return default


def targz(output_filename, source_dir):
    with tarfile.open(output_filename, "w:gz") as tar:
        tar.add(source_dir, arcname=os.path.basename(source_dir))
        
        
def timestamp():
  now_time = datetime.now()
  if PY3:
    return now_time.timestamp()
  else:
    epoch = datetime.utcfromtimestamp(0)
    total_seconds = (now_time - epoch).total_seconds()
    # total_seconds will be in decimals (millisecond precision)
    return total_seconds


# -*- coding: UTF-8 -*-
# @Time    : 17-9-19
# @File    : deep_analysis.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import numpy as np
from antgo.utils._resize import *
from antgo.utils.encode import *
import base64
import copy

# TODO: update incoming (data relevant)
def _whats_data(data_source, id, infos):
  d = data_source.at(id)
  if type(d) == tuple or type(d) == list:
    d = d[0]
  
  if data_source.dataset_type == "TEXT":
    # text
    return {'type': 'TEXT', 'data': d}
  elif data_source.dataset_type == "IMAGE":
    # image
    png_data = png_encode(d, True)
    return {'type': 'IMAGE', 'data': base64.b64encode(png_data).decode('utf-8')}
  
  return None

def discrete_multi_model_measure_analysis(samples_score, samples_map, data_source, filter_tag=None, random_sampling=5):
  # 95%, 52%, 42%, 13%, only best, 0%
  # correct is 1; error is 0
  # 95% - sample could be recognized correctly by 95% models
  # 52% - sample could be recognized correctly by 52% models
  # ...
  # filter by tag
  max_num = samples_score.shape[1]
  assert(max_num == len(samples_map))

  remained_id = []
  if filter_tag is not None:
    for data_index in range(max_num):
      _, label = data_source.at(samples_map[data_index]['id'])
      if filter_tag in label['tag']:
        remained_id.append(data_index)

  if len(remained_id) == 0:
    remained_id = range(max_num)

  samples_score = samples_score[:, remained_id]
  model_num, samples_num = samples_score.shape[0:2]

  # sort cols
  cols_scores = np.sum(samples_score, axis=0)
  cols_orders = np.argsort(-cols_scores)    # decending
  ordered_samples_score = samples_score[:, cols_orders]
  ordered_data_id = [samples_map[remained_id[i]]['id'] for i in cols_orders]

  # sort rows
  rows_scores = np.sum(ordered_samples_score, axis=1)
  rows_orders = np.argsort(-rows_scores)    # decending
  ordered_samples_score = ordered_samples_score[rows_orders, :]
  ordered_model_id = [i for i in rows_orders]

  score_hist = np.sum(ordered_samples_score, axis=0)
  score_hist = score_hist / float(model_num)
  # 95% (94% ~ 96%)
  pos_95 = np.searchsorted(-score_hist, [-0.95])[0]
  pos_95_s = int(np.maximum(0, pos_95 - random_sampling/2))
  pos_95_e = int(np.minimum(samples_num, pos_95_s+random_sampling))

  if pos_95_e <= pos_95_s:
    pos_95_e = pos_95_s + 1
  region_95 = cols_orders[pos_95_s: pos_95_e]
  region_95 = [_whats_data(data_source,
                           samples_map[remained_id[index]]['id'],
                           samples_map[remained_id[index]]) for index in region_95]

  # 52%
  pos_52 = np.searchsorted(-score_hist, [-0.52])[0]
  pos_52_s = int(np.maximum(0, pos_52 - random_sampling/2))
  pos_52_e = int(np.minimum(samples_num, pos_52_s+random_sampling))
  if pos_52_e <= pos_52_s:
    pos_52_e = pos_95_s + 1
  region_52 = cols_orders[pos_52_s:pos_52_e]
  region_52 = [_whats_data(data_source,
                           samples_map[remained_id[index]]['id'],
                           samples_map[remained_id[index]]) for index in region_52]

  # 42%
  pos_42 = np.searchsorted(-score_hist, [-0.42])[0]
  pos_42_s = int(np.maximum(0, pos_42 - random_sampling/2))
  pos_42_e = int(np.minimum(samples_num, pos_42_s+random_sampling))
  if pos_42_e <= pos_42_s:
    pos_42_e = pos_42_s + 1
  # region_42 = np.arange(pos_42_s, pos_42_e)
  region_42 = cols_orders[pos_42_s:pos_42_e]
  region_42 = [_whats_data(data_source,
                           samples_map[remained_id[index]]['id'],
                           samples_map[remained_id[index]]) for index in region_42]
  
  # 13%
  pos_13 = np.searchsorted(-score_hist, [-0.13])[0]
  pos_13_s = int(np.maximum(0, pos_13 - random_sampling/2))
  pos_13_e = int(np.minimum(samples_num, pos_13_s+random_sampling))
  if pos_13_e <= pos_13_s:
    pos_13_e = pos_13_s + 1
  region_13 = cols_orders[pos_13_s: pos_13_e]
  # region_13 = np.arange(pos_13_s, pos_13_e)
  region_13 = [_whats_data(data_source,
                           samples_map[remained_id[index]]['id'],
                           samples_map[remained_id[index]]) for index in region_13]

  # only best
  pos_one = np.searchsorted(-score_hist, [-1.0])[0]
  pos_one_s = int(np.maximum(0, pos_one - random_sampling/2))
  pos_one_e = int(np.minimum(samples_num, pos_one_s+random_sampling))
  if pos_one_e <= pos_one_s:
    pos_one_e = pos_one_s + 1
  # region_one = np.arange(pos_one_s,pos_one_e)
  region_one = cols_orders[pos_one_s: pos_one_e]
  region_one = [_whats_data(data_source,
                            samples_map[remained_id[index]]['id'],
                            samples_map[remained_id[index]]) for index in region_one]

  # 0%
  pos_zero = np.searchsorted(-score_hist, [0.0])[0]
  pos_zero_s = int(np.maximum(0, pos_zero - random_sampling / 2))
  pos_zero_e = int(np.minimum(samples_num, pos_zero_s + random_sampling))
  if pos_zero_e <= pos_zero_s:
    pos_zero_e = pos_zero_s + 1
  region_zero = cols_orders[pos_zero_s: pos_zero_e]
  # region_zero = np.arange(pos_zero_s, pos_zero_e)
  region_zero = [_whats_data(data_source,
                            samples_map[remained_id[index]]['id'],
                            samples_map[remained_id[index]]) for index in region_zero]

  return ordered_samples_score.tolist(), \
         ordered_model_id, \
         ordered_data_id, \
         region_95, \
         region_52, \
         region_42, \
         region_13, \
         region_one, \
         region_zero


def continuous_multi_model_measure_analysis(samples_score, samples_map, data_source, filter_tag=None, random_sampling=10):
  max_num = samples_score.shape[1]
  assert(max_num == len(samples_map))
  # filter by tag
  remained_id = []
  if filter_tag is not None:
    for data_index in range(max_num):
      _, label = data_source.at(samples_map[data_index]['id'])
      if filter_tag == label['tag']:
        remained_id.append(data_index)
  
  if len(remained_id) == 0:
    remained_id = range(max_num)
  
  samples_score = samples_score[:, remained_id]
  
  # low,  middle,  high
  model_num, samples_num = samples_score.shape[0:2]

  # reorder cols (samples)
  model_score = np.sum(samples_score, axis=0)
  reorganized_sample_id = np.argsort(-model_score)    # decending
  output_reorganized_sample_id = [samples_map[remained_id[i]]['id'] for i in reorganized_sample_id]
  reorganized_samples_score = samples_score[:, reorganized_sample_id]
  samples_score = reorganized_samples_score.copy()

  # reorder rows (models)
  s = np.sum(reorganized_samples_score, axis=1)
  reorganized_model_id = np.argsort(-s)               # decending
  reorganized_samples_score = samples_score[reorganized_model_id, :]

  # high score region (0 ~ 1/10) - good
  region_start = 0
  region_end = int(np.minimum(samples_num / 10 * 1, samples_num))
  if region_end == 0:
    region_end = 1
  high_region_sampling = reorganized_sample_id[region_start:region_end]
  high_region_random_sampling = np.minimum(len(high_region_sampling), random_sampling)
  high_region_sampling = np.random.choice(high_region_sampling, high_region_random_sampling, False)
  high_region_sampling = [_whats_data(data_source,
                                      samples_map[remained_id[index]]['id'],
                                      samples_map[remained_id[index]]) for index in high_region_sampling]
  
  # middle score region (4/10 ~ 6/10) - just so so
  region_start = int(np.maximum(samples_num / 10 * 4, 0))
  region_end = int(np.minimum(samples_num / 10 * 6, samples_num))
  if region_end == 0:
    region_end = 1
  middle_region_sampling = reorganized_sample_id[region_start: region_end]
  middle_region_random_sampling = np.minimum(len(middle_region_sampling), random_sampling)
  middle_region_sampling = np.random.choice(middle_region_sampling, middle_region_random_sampling, False)
  middle_region_sampling = [_whats_data(data_source,
                                        samples_map[remained_id[index]]['id'],
                                        samples_map[remained_id[index]]) for index in middle_region_sampling]
  
  # low score region (9/10 ~ 10/10) - bad
  region_start = int(np.maximum(samples_num / 10 * 9, 0))
  region_end = int(np.minimum(samples_num / 10 * 10, samples_num))
  if region_end == 0:
    region_end = 1
  
  low_region_sampling = reorganized_sample_id[region_start: region_end]
  low_region_random_sampling = np.minimum(len(low_region_sampling), random_sampling)
  low_region_sampling = np.random.choice(low_region_sampling, low_region_random_sampling, False)
  low_region_sampling = [_whats_data(data_source,
                                     samples_map[remained_id[index]]['id'],
                                     samples_map[remained_id[index]]) for index in low_region_sampling]

  return reorganized_samples_score, \
         reorganized_model_id, \
         output_reorganized_sample_id, \
         low_region_sampling, \
         middle_region_sampling, \
         high_region_sampling


import numpy as np
cimport numpy as np
cdef _uint8_resize(unsigned char* data, src_size, target_size):
    cdef unsigned int height = src_size[0]
    cdef unsigned int width = src_size[1]
    cdef unsigned int channels = src_size[2]

    cdef unsigned int target_height = target_size[0]
    cdef unsigned int target_width = target_size[1]
    cdef np.ndarray[np.uint8_t, ndim=1] target_image = np.zeros((target_height * target_width * channels), dtype=np.uint8)
    cdef unsigned char* target_data = <unsigned char*> target_image.data

    cdef float scale_x =  float(width) / float(target_width)
    cdef float scale_y = float(height) / float(target_height)

    cdef unsigned int y = 0
    cdef unsigned int x = 0
    cdef float fy = 0.0
    cdef int sy = 0
    cdef float fx = 0.0
    cdef int sx = 0

    cdef unsigned int c = 0
    cdef int src_width_size = width * channels
    cdef int target_width_size = target_width * channels

    for y in range(target_height):
        fy = <float>((y + 0.5) * scale_y - 0.5)
        sy = <int> fy
        fy -= sy

        if fy < 0:
            fy = 0

        if sy < 0:
            fy = 0
            sy = 0
        if sy >= height -1:
            fy = 1
            sy = height -2

        for x in range(target_width):
            fx = <float>((x + 0.5) * scale_x - 0.5)
            sx = <int> fx
            fx -= sx

            if fx < 0:
                fx = 0

            if sx < 0:
                fx = 0
                sx = 0

            if sx >= width - 1:
                fx = 1
                sx = width - 2

            for c in range(channels):
                target_data[y * target_width_size + x * channels + c] = <unsigned char>((1.0 - fx) * (1.0 - fy) * data[sy*src_width_size+ sx*channels + c] + \
                                                                         (1.0 - fx) * fy * data[(sy+1) * src_width_size + sx*channels + c] + \
                                                                         fx * (1.0 - fy) * data[sy*src_width_size + (sx+1)*channels + c] + \
                                                                         fx * fy * data[(sy+1)*src_width_size + (sx+1)*channels + c] + 0.5)
    return target_image

cdef _f32_resize(float* data, src_size, target_size):
    cdef unsigned int height = src_size[0]
    cdef unsigned int width = src_size[1]
    cdef unsigned int channels = src_size[2]

    cdef unsigned int target_height = target_size[0]
    cdef unsigned int target_width = target_size[1]
    cdef np.ndarray[np.float32_t, ndim=1] target_image = np.zeros((target_height * target_width * channels), dtype=np.float32)
    cdef float* target_data = <float*> target_image.data

    cdef float scale_x =  float(width) / float(target_width)
    cdef float scale_y = float(height) / float(target_height)

    cdef unsigned int y = 0
    cdef unsigned int x = 0
    cdef float fy = 0.0
    cdef int sy = 0
    cdef float fx = 0.0
    cdef int sx = 0

    cdef unsigned int c = 0
    cdef int src_width_size = width * channels
    cdef int target_width_size = target_width * channels

    for y in range(target_height):
        fy = <float>((y + 0.5) * scale_y - 0.5)
        sy = <int> fy
        fy -= sy

        if fy < 0:
            fy = 0

        if sy < 0:
            fy = 0
            sy = 0
        if sy >= height -1:
            fy = 1
            sy = height -2

        for x in range(target_width):
            fx = <float>((x + 0.5) * scale_x - 0.5)
            sx = <int> fx
            fx -= sx

            if fx < 0:
                fx = 0

            if sx < 0:
                fx = 0
                sx = 0

            if sx >= width - 1:
                fx = 1
                sx = width - 2

            for c in range(channels):
                target_data[y * target_width_size + x * channels + c] = (1.0 - fx) * (1.0 - fy) * data[sy*src_width_size+ sx*channels + c] + \
                                                                         (1.0 - fx) * fy * data[(sy+1) * src_width_size + sx*channels + c] + \
                                                                         fx * (1.0 - fy) * data[sy*src_width_size + (sx+1)*channels + c] + \
                                                                         fx * fy * data[(sy+1)*src_width_size + (sx+1)*channels + c]
    return target_image


cdef _d64_resize(double* data, src_size, target_size):
    cdef unsigned int height = src_size[0]
    cdef unsigned int width = src_size[1]
    cdef unsigned int channels = src_size[2]

    cdef unsigned int target_height = target_size[0]
    cdef unsigned int target_width = target_size[1]
    cdef np.ndarray[np.double_t, ndim=1] target_image = np.zeros((target_height * target_width * channels), dtype=np.double)
    cdef double* target_data = <double*> target_image.data

    cdef float scale_x =  float(width) / float(target_width)
    cdef float scale_y = float(height) / float(target_height)

    cdef unsigned int y = 0
    cdef unsigned int x = 0
    cdef float fy = 0.0
    cdef int sy = 0
    cdef float fx = 0.0
    cdef int sx = 0

    cdef unsigned int c = 0
    cdef unsigned int src_width_size = width * channels
    cdef unsigned int target_width_size = target_width * channels

    for y in range(target_height):
        fy = <float>((y + 0.5) * scale_y - 0.5)
        sy = <int> fy
        fy -= sy

        if fy < 0:
            fy = 0

        if sy < 0:
            fy = 0
            sy = 0
        if sy >= height -1:
            fy = 1
            sy = height -2

        for x in range(target_width):
            fx = <float>((x + 0.5) * scale_x - 0.5)
            sx = <int> fx
            fx -= sx

            if fx < 0:
                fx = 0

            if sx < 0:
                fx = 0
                sx = 0

            if sx >= width - 1:
                fx = 1
                sx = width - 2

            for c in range(channels):
                target_data[y * target_width_size + x * channels + c] = (1.0 - fx) * (1.0 - fy) * data[sy*src_width_size+ sx*channels + c] + \
                                                                         (1.0 - fx) * fy * data[(sy+1) * src_width_size + sx*channels + c] + \
                                                                         fx * (1.0 - fy) * data[sy*src_width_size + (sx+1)*channels + c] + \
                                                                         fx * fy * data[(sy+1)*src_width_size + (sx+1)*channels + c]
    return target_image


def resize(image, target):
    cdef unsigned int height = image.shape[0]
    cdef unsigned int width = image.shape[1]
    cdef unsigned int channels = 1 if len(image.shape) == 2 else image.shape[2]
    cdef is_ndim_2 = 1 if len(image.shape) == 2 else 0
    cdef np.ndarray[np.double_t, ndim=1] double_image
    cdef np.ndarray[np.float32_t, ndim=1] float_image
    cdef np.ndarray[np.uint8_t, ndim=1] uint8_image

    if image.dtype == np.float or image.dtype == np.double or image.dtype == np.float64:
        double_image = np.ascontiguousarray(image.ravel(), dtype=np.double)
        target_image = _d64_resize(<double*>double_image.data, [height, width, channels], target)
        if is_ndim_2 == 1:
            return target_image.reshape(target)
        else:
            return target_image.reshape((target[0], target[1], channels))
    elif image.dtype == np.float32:
        float_image = np.ascontiguousarray(image.ravel(), dtype=np.float32)
        target_image = _f32_resize(<float*>float_image.data, [height, width, channels], target)
        if is_ndim_2 == 1:
            return target_image.reshape(target)
        else:
            return target_image.reshape((target[0], target[1], channels))
    if image.dtype == np.uint8:
        uint8_image = np.ascontiguousarray(image.ravel(), dtype=np.uint8)
        target_image = _uint8_resize(<unsigned char*>uint8_image.data, [height, width, channels], target)
        if is_ndim_2 == 1:
            return target_image.reshape(target)
        else:
            return target_image.reshape((target[0], target[1], channels))
    else:
        raise TypeError



# -*- coding: UTF-8 -*-
# @Time : 17/03/2018
# @File : utils.py
# @Author: Jian <jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import imghdr
import os
import shutil

def url_path_join(*pieces):
  """Join components of url into a relative url

  Use to prevent double slash when joining subpath. This will leave the
  initial and final / in place

  Copied from notebook.utils.url_path_join
  """
  initial = pieces[0].startswith('/')
  final = pieces[-1].endswith('/')
  stripped = [s.strip('/') for s in pieces]
  result = '/'.join(s for s in stripped if s)

  if initial:
    result = '/' + result
  if final:
    result = result + '/'
  if result == '//':
    result = '/'

  return result



# -*- coding: UTF-8 -*-
# @Time    : 18-3-26
# @File    : repeat_statistic.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import numpy as np
import copy
from antgo.resource.html import *
from antgo.measures.significance import *

def multi_repeats_measures_statistic(multi_statistics, method='repeated-holdout'):
  # time
  series_elapsed_time = []
  series_elapsed_time_per_sample = []

  # cpu
  series_cpu_mean_usage = []
  series_cpu_median_usage = []
  series_cpu_max_usage = []

  # memory
  series_mem_mean_usage = []
  series_mem_median_usage = []
  series_mem_max_usage = []

  # measures
  series_measures = []
  cpu_model = None
  title = ""

  for statistic_result_index, statistic_result in enumerate(multi_statistics):
    for ant_name, ant_statistic in statistic_result.items():
      title = ant_name
      if 'time' in ant_statistic:
        series_elapsed_time.append(ant_statistic['time']['elapsed_time'])
        if 'elapsed_time_per_sample' in ant_statistic['time']:
          series_elapsed_time_per_sample.append(ant_statistic['time']['elapsed_time_per_sample'])

      if 'cpu' in ant_statistic:
        if 'cpu_mean_usage' in ant_statistic['cpu']:
          series_cpu_mean_usage.append(ant_statistic['cpu']['cpu_mean_usage'])
        if 'cpu_median_usage' in ant_statistic['cpu']:
          series_cpu_median_usage.append(ant_statistic['cpu']['cpu_median_usage'])
        if 'cpu_max_usage' in ant_statistic['cpu']:
          series_cpu_max_usage.append(ant_statistic['cpu']['cpu_max_usage'])

        if 'mem_mean_usage' in ant_statistic['cpu']:
          series_mem_mean_usage.append(ant_statistic['cpu']['mem_mean_usage'])
        if 'mem_median_usage' in ant_statistic['cpu']:
          series_mem_median_usage.append(ant_statistic['cpu']['mem_median_usage'])
        if 'mem_max_usage' in ant_statistic['cpu']:
          series_mem_max_usage.append(ant_statistic['cpu']['mem_max_usage'])
        if 'cpu_model' in ant_statistic['cpu']:
          cpu_model = ant_statistic['cpu']['cpu_model']

      if 'gpu' in ant_statistic:
        pass

      if 'measure' in ant_statistic:
        for measure_index, measure_statistic in enumerate(ant_statistic['measure']):
          if statistic_result_index == 0:
            series_measures.append(copy.deepcopy(measure_statistic))

            for per_value in series_measures[measure_index]['statistic']['value']:
              if per_value['type'] == 'SCALAR':
                # (1) shape - scalar
                # (2) shape - [scalar, scalar, ...]
                per_value['value'] = [per_value['value']]
              elif per_value['type'] == 'CURVE':
                # (1) shape - [np.array.tolist(),np.array.tolist(),...]
                per_value['value'] = [per_value['value']]
              elif per_value['type'] == 'MATRIX':
                # (1) shape - Matrix(np.array)
                per_value['value'] = [per_value['value']]
            #
            continue

          # combine others
          final_measure_statistic = series_measures[measure_index]
          for per_value_index, per_value in enumerate(measure_statistic['statistic']['value']):
            if per_value['type'] == 'SCALAR':
              # (1) shape - scalar
              # (2) shape - [scalar, scalar, ...]
              final_measure_statistic['statistic']['value'][per_value_index]['value'].append(per_value['value'])
            elif per_value['type'] == 'CURVE':
              # (1) shape - [np.array,np.array,...]
              final_measure_statistic['statistic']['value'][per_value_index]['value'].append([a for a in per_value['value']])
            elif per_value['type'] == 'MATRIX':
              # (1) shape - Matrix
              final_measure_statistic['statistic']['value'][per_value_index]['value'].append(per_value['value'])


  multi_statistics = {}
  # time
  if len(series_elapsed_time) > 0:
    multi_statistics['time'] = {}
    series_elapsed_time_mean = np.mean(series_elapsed_time)
    series_elapsed_time_std = np.std(series_elapsed_time)
    multi_statistics['time']['elapsed_time'] = float(series_elapsed_time_mean)
    multi_statistics['time']['elapsed_time_interval'] = float(series_elapsed_time_std)

  if len(series_elapsed_time_per_sample) > 0:
    series_elapsed_time_mean_per_sample = float(np.mean(series_elapsed_time_per_sample))
    series_elapsed_time_std_per_sample = float(np.std(series_elapsed_time_per_sample))
    multi_statistics['time']['elapsed_time_per_sample'] = series_elapsed_time_mean_per_sample
    multi_statistics['time']['elapsed_time_per_sample_interval'] = series_elapsed_time_std_per_sample

  # cpu
  if cpu_model is not None:
    multi_statistics['cpu'] = {}
    multi_statistics['cpu']['cpu_model'] = cpu_model

    series_cpu_mean_usage_mean = np.mean(series_cpu_mean_usage)
    series_cpu_mean_usage_std = np.std(series_cpu_mean_usage)
    multi_statistics['cpu']['cpu_mean_usage'] = float(series_cpu_mean_usage_mean)
    multi_statistics['cpu']['cpu_mean_usage_interval'] = float(series_cpu_mean_usage_std)

    series_cpu_median_usage_mean = np.mean(series_cpu_median_usage)
    series_cpu_median_usage_std = np.std(series_cpu_median_usage)
    multi_statistics['cpu']['cpu_median_usage'] = float(series_cpu_median_usage_mean)
    multi_statistics['cpu']['cpu_median_usage_interval'] = float(series_cpu_median_usage_std)

    series_cpu_max_usage_mean = np.mean(series_cpu_max_usage)
    series_cpu_max_usage_std = np.std(series_cpu_max_usage)
    multi_statistics['cpu']['cpu_max_usage'] = float(series_cpu_max_usage_mean)
    multi_statistics['cpu']['cpu_max_usage_interval'] = float(series_cpu_max_usage_std)

    # memory
    series_mem_mean_usage_mean = np.mean(series_mem_mean_usage)
    series_mem_mean_usage_std = np.std(series_mem_mean_usage)
    multi_statistics['cpu']['mem_mean_usage'] = float(series_mem_mean_usage_mean)
    multi_statistics['cpu']['mem_mean_usage_interval'] = float(series_mem_mean_usage_std)

    series_mem_median_usage_mean = np.mean(series_mem_median_usage)
    series_mem_median_usage_std = np.std(series_mem_median_usage)
    multi_statistics['cpu']['mem_median_usage'] = float(series_mem_median_usage_mean)
    multi_statistics['cpu']['mem_median_usage_interval'] = float(series_mem_median_usage_std)

    series_mem_max_usage_mean = np.mean(series_mem_max_usage)
    series_mem_max_usage_std = np.std(series_mem_max_usage)
    multi_statistics['cpu']['mem_max_usage'] = float(series_mem_max_usage_mean)
    multi_statistics['cpu']['mem_max_usage_interval'] = float(series_mem_max_usage_std)

  # measure
  for measure in series_measures:
    for per_value in measure['statistic']['value']:
      if per_value['type'] == 'SCALAR':
        scalar_array = np.array(per_value['value'])
        if len(scalar_array.shape) == 1:
          per_value['value'] = float(np.mean(scalar_array))
          if method == 'bootstrap':
            per_value['interval'] = bootstrap_direct_confidence_interval(scalar_array)
          else:
            per_value['interval'] = float(np.std(scalar_array))
        else:
          assert(len(scalar_array.shape) == 2)
          per_value['value'] = np.mean(scalar_array, axis=0).tolist()
          per_value['interval'] = np.std(scalar_array, axis=0).tolist()
      elif per_value['type'] == 'CURVE':
        # shape (N,curves_num,point_num,2)
        scalar_array = np.array(per_value['value'])
        assert(len(scalar_array.shape) == 4)
        x, y = np.split(scalar_array, 2, axis=3)
        x = np.squeeze(x, axis=3)
        xt = np.transpose(x, [1, 2, 0])
        x_mean = np.mean(xt, axis=2)

        y = np.squeeze(y, axis=3)
        yt = np.transpose(y, [1, 2, 0])
        y_mean = np.mean(yt, axis=2) # (curves_num,point_num)
        y_std = np.std(yt, axis=2)   # (curves_num,point_num)

        # shape (curves_num,point_num,2)
        xy = np.concatenate((np.expand_dims(x_mean, 2), np.expand_dims(y_mean, 2)), axis=2)
        curves_num = xy.shape[0]
        curves_list = np.split(xy, curves_num, axis=0)

        xy_std = np.concatenate((np.expand_dims(x_mean,2),np.expand_dims(y_std,2)),axis=2)
        curves_std_list = np.split(xy_std,curves_num,axis=0)

        per_value['value'] = [np.squeeze(cl).tolist() for cl in curves_list]
        per_value['interval'] = [np.squeeze(cl).tolist() for cl in curves_std_list]
      elif per_value['type'] == 'MATRIX':
        # shape (N,rows,cols)
        scalar_array = np.array(per_value['value'])
        assert(len(scalar_array.shape) == 3)
        m_mean = np.mean(scalar_array, axis=0)
        m_std = np.std(scalar_array, axis=0)

        per_value['value'] = m_mean.tolist()
        per_value['interval'] = m_std.tolist()

  multi_statistics['measure'] = series_measures
  ant_statistic_warp = {}
  ant_statistic_warp[title] = multi_statistics
  return ant_statistic_warp

if __name__ == '__main__':
    # experiment 1
    experiment_1_statis = {}
    experiment_1_statis['aa'] = {}
    experiment_1_statis['aa']['time'] = {}
    experiment_1_statis['aa']['time']['elapsed_time'] = 1.2
    experiment_1_statis['aa']['cpu'] = {}
    experiment_1_statis['aa']['cpu']['mem_mean_usage'] = 23
    experiment_1_statis['aa']['cpu']['mem_median_usage'] = 12
    experiment_1_statis['aa']['cpu']['mem_max_usage'] = 44
    experiment_1_statis['aa']['cpu']['cpu_mean_usage'] = 55
    experiment_1_statis['aa']['cpu']['cpu_median_usage'] = 11
    experiment_1_statis['aa']['cpu']['cpu_max_usage'] = 22
    experiment_1_statis['aa']['cpu']['cpu_model'] = 'aabbcc'

    voc_measure = {'statistic':{'name':'voc',
                             'value':[{'name':'MAP','value':[23.0,11.0,12.0],'type':'SCALAR','x':'class','y':'Mean Average Precision'},
                                      {'name':'Mean-MAP','value':0.23,'type':'SCALAR'}]}}

    roc_auc_measure = {'statistic':{'name':'roc_auc',
                             'value':[{'name':'ROC','value':[(np.ones((3,2)) * 3).tolist(),(np.ones((3,2))*2).tolist()],'type':'CURVE','x':'FP','y':'TP'},
                                      {'name':'AUC','value':[0.3,0.4],'type':'SCALAR','x':'class','y':'AUC'}]}}

    pr_f1_measure = {'statistic':{'name':'pr_f1',
                             'value':[{'name':'Precision-Recall','value':[(np.ones((4,2))*4).tolist(),(np.ones((4,2))*3).tolist()],'type':'CURVE','x':'precision','y':'recall'},
                                      {'name':'F1','value':[1.0,2.0],'type':'SCALAR','x':'class','y':'F1'}]}}

    confusion_m = {'statistic': {'name': 'cm',
                              'value': [{'name':'ccmm','value':(np.ones((3,4))*3).tolist(),'type':'MATRIX','x':'class','y':'class'}]}}

    experiment_1_statis['aa']['measure'] = [voc_measure,roc_auc_measure,pr_f1_measure,confusion_m]
    # experiment 2
    experiment_2_statis = {}
    experiment_2_statis['aa'] = {}
    experiment_2_statis['aa']['time'] = {}
    experiment_2_statis['aa']['time']['elapsed_time'] = 1.2
    experiment_2_statis['aa']['cpu'] = {}
    experiment_2_statis['aa']['cpu']['mem_mean_usage'] = 23
    experiment_2_statis['aa']['cpu']['mem_median_usage'] = 12
    experiment_2_statis['aa']['cpu']['mem_max_usage'] = 44
    experiment_2_statis['aa']['cpu']['cpu_mean_usage'] = 55
    experiment_2_statis['aa']['cpu']['cpu_median_usage'] = 11
    experiment_2_statis['aa']['cpu']['cpu_max_usage'] = 22
    experiment_2_statis['aa']['cpu']['cpu_model'] = 'aabbcc'

    voc_measure = {'statistic': {'name': 'voc',
                                 'value': [{'name': 'MAP', 'value': [2.0, 1.0, 2.0], 'type': 'SCALAR', 'x': 'class',
                                            'y': 'Mean Average Precision'},
                                           {'name': 'Mean-MAP', 'value': 0.12, 'type': 'SCALAR'}]}}

    roc_auc_measure = {'statistic': {'name': 'roc_auc',
                                     'value': [
                                         {'name': 'ROC', 'value': [(np.ones((3, 2))*4).tolist(), (np.ones((3, 2)) * 1).tolist()], 'type': 'CURVE',
                                          'x': 'FP', 'y': 'TP'},
                                         {'name': 'AUC', 'value': [0.12, 0.34], 'type': 'SCALAR', 'x': 'class',
                                          'y': 'AUC'}]}}

    pr_f1_measure = {'statistic': {'name': 'pr_f1',
                                   'value': [
                                       {'name': 'Precision-Recall', 'value': [(np.ones((4, 2)) * 1).tolist(), (np.ones((4, 2)) * 2).tolist()],
                                        'type': 'CURVE', 'x': 'precision', 'y': 'recall'},
                                       {'name': 'F1', 'value': [1.4, 1.0], 'type': 'SCALAR', 'x': 'class', 'y': 'F1'}]}}

    confusion_m = {'statistic': {'name': 'cm',
                                 'value': [{'name': 'ccmm', 'value': (np.ones((3, 4)) * 8).tolist(), 'type': 'MATRIX',
                                            'x': 'class', 'y': 'class'}]}}

    experiment_2_statis['aa']['measure'] = [voc_measure, roc_auc_measure, pr_f1_measure,confusion_m]

    ss = multi_repeats_measures_statistic([experiment_1_statis,experiment_2_statis])
    everything_to_html(ss,'/home/mi')
    print(ss)


# encoding=utf-8
# @Time    : 17-5-24
# @File    : processbar.py
# @Author  :
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import sys


def progress_bar(progress, callback_func=None, prefix="", bar_length=50):
    status = ""
    if isinstance(progress, int):
        progress = float(progress)

    if not isinstance(progress, float):
        progress = 0
        status = 'Error: Progress Var Must be Float\r\n'
    if progress < 0:
        progress = 0
        status = "Halt ...\r\n"
        return
    if progress >= 1.0:
        progress = 1.0
        status = '\n'

    block = int(round(bar_length * progress))
    text = "\r%s[ %s] %.2f%% %s" % (prefix, "=" * block +">" + " " * int(bar_length - block), progress * 100, status)
    sys.stdout.write(text)
    sys.stdout.flush()

    if progress >= 1.0:
        if callback_func is not None:
            callback_func()


def iter_progress_bar(iter_at,iter_intervals,bar_prefix="",info=None,info_format=None,bar_length=50):
    iter_at = int(iter_at)
    iter_intervals = int(iter_intervals)
    progress = (iter_at % iter_intervals + 1.0) / float(iter_intervals)

    callback_func = None
    if info is not None:
        if info_format is None:
            info_format = ','.join(['%0.2f' for _ in range(len(info))])
        callback_func = lambda : print(info_format%tuple([_() for _ in info]))

    progress_bar(progress, callback_func, bar_prefix, bar_length)


# -*- coding: UTF-8 -*-
# @Time    : 2018/11/21 5:14 PM
# @File    : stublayers.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.automl.stublayer import *
from antgo.automl.basestublayers import *
import tensorflow as tf
import tensorflow.contrib.slim as slim
import functools
import numpy as np


class StubDense(BaseStubDense):
  def __init__(self, input_units, units, input=None, output=None, **kwargs):
    super(StubDense, self).__init__(input_units, units, input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #       with tf.variable_scope(self.cell_name, 'cell'):
    #         return functools.partial(slim.fully_connected,num_outputs=self.units)(*args, **kwargs)
    # else:
    return functools.partial(slim.fully_connected, num_outputs=self.units)(*args, **kwargs)


class StubConv2d(BaseStubConv2d):
  def __init__(self, input_channel, filters, kernel_size_h, kernel_size_w, rate_h=1, rate_w=1, stride=1, input=None, output=None, **kwargs):
    super(StubConv2d, self).__init__(input_channel,
                                     filters,
                                     kernel_size_h,
                                     kernel_size_w,
                                     rate_h,
                                     rate_w,
                                     stride,
                                     input,
                                     output,**kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #       with tf.variable_scope(self.cell_name, 'cell'):
    #         return functools.partial(slim.conv2d,
    #                                  num_outputs=self.filters,
    #                                  kernel_size=[self.kernel_size_h,
    #                                               self.kernel_size_w],
    #                                  rate=[self.rate_h, self.rate_w],
    #                                  stride=self.stride)(*args, **kwargs)
    #
    # else:
    return functools.partial(slim.conv2d,
                             num_outputs=self.filters,
                             kernel_size=[self.kernel_size_h,
                                          self.kernel_size_w],
                             rate=[self.rate_h, self.rate_w],
                             stride=self.stride)(*args, **kwargs)


class StubSeparableConv2d(BaseStubSeparableConv2d):
  def __init__(self, input_channel, filters, kernel_size_h, kernel_size_w, rate_h=1, rate_w=1, stride=1, input=None, output=None, **kwargs):
    super(StubSeparableConv2d, self).__init__(input_channel,
                                              filters,
                                              kernel_size_h,
                                              kernel_size_w,
                                              rate_h,
                                              rate_w,
                                              stride,
                                              input,
                                              output,**kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #       with tf.variable_scope(self.cell_name, 'cell'):
    #         return functools.partial(slim.separable_conv2d,
    #                           num_outputs=self.filters,
    #                           kernel_size=[self.kernel_size_h, self.kernel_size_w],
    #                           rate=[self.rate_h, self.rate_w],
    #                           depth_multiplier=1)(*args, **kwargs)
    # else:
    return functools.partial(slim.separable_conv2d,
                      num_outputs=self.filters,
                      kernel_size=[self.kernel_size_h, self.kernel_size_w],
                      rate=[self.rate_h, self.rate_w],
                      depth_multiplier=1)(*args, **kwargs)


class StubSPP(BaseStubSPP):
  def __init__(self, grid_h, grid_w, input=None, output=None, **kwargs):
    super(StubSPP, self).__init__(grid_h, grid_w, input, output,**kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #       with tf.variable_scope(self.cell_name, 'cell'):
    #         # 1.step average pooling
    #         _, input_h, input_w, input_c = self.input.shape
    #         grid_h = min(self.grid_h, input_h)
    #         grid_w = min(self.grid_w, input_w)
    #         output = functools.partial(slim.avg_pool2d,
    #                                    kernel_size=[grid_h, grid_w],
    #                                    stride=[grid_h, grid_w])(*args, **kwargs)
    #
    #         # 2.step 1x1 convolution
    #         output = slim.conv2d(output, input_c, 1)
    #
    #         # 3.step bilinear resize
    #         output = tf.image.resize_bilinear(output, [input_h, input_w])
    #
    #         return output
    # else:
    # 1.step average pooling
    # _, input_h, input_w, input_c = self.input.shape
    _, input_h, input_w, input_c = args[0].shape
    grid_h = min(self.grid_h, input_h)
    grid_w = min(self.grid_w, input_w)
    output = functools.partial(slim.avg_pool2d,
                               kernel_size=[grid_h, grid_w],
                               stride=[grid_h, grid_w])(*args, **kwargs)

    # 2.step 1x1 convolution
    output = slim.conv2d(output, input_c, 1)

    # 3.step bilinear resize
    output = tf.image.resize_bilinear(output, [input_h, input_w])

    return output


class StubConcatenate(BaseStubConcatenate):
  def __init__(self, input=None, output=None, **kwargs):
    super(StubConcatenate, self).__init__(input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #       with tf.variable_scope(self.cell_name, 'cell'):
    #         return functools.partial(tf.concat, axis=-1)(*args, **kwargs)
    # else:
    return functools.partial(tf.concat, axis=-1)(*args, **kwargs)


class StubAdd(BaseStubAdd):
  def __init__(self, input=None, output=None, **kwargs):
    super(StubAdd, self).__init__(input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #       with tf.variable_scope(self.cell_name, 'cell'):
    #         return functools.partial(tf.add)(*args, **kwargs)
    # else:
    return functools.partial(tf.add)(args[0][0],args[0][1], **kwargs)


class StubDot(BaseStubDot):
  def __init__(self, input=None, output=None, **kwargs):
    super(StubDot, self).__init__(input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #       with tf.variable_scope(self.cell_name, 'cell'):
    #         return functools.partial(tf.add)(*args, **kwargs)
    # else:
    return functools.partial(tf.multiply)(args[0][0], args[0][1], **kwargs)



class StubFlatten(BaseStubFlatten):
  def __init__(self, input=None, output=None, **kwargs):
    super(StubFlatten, self).__init__(input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #     with tf.variable_scope(self.cell_name, 'cell'):
    #       return functools.partial(slim.flatten)(*args, **kwargs)
    # else:
    return functools.partial(slim.flatten)(*args, **kwargs)


class StubReLU(BaseStubReLU):
  def __init__(self, input=None, output=None, **kwargs):
    super(StubReLU, self).__init__(input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #     with tf.variable_scope(self.cell_name, 'cell'):
    #       return functools.partial(tf.nn.relu)(*args, **kwargs)
    # else:
    return functools.partial(tf.nn.relu)(*args, **kwargs)


class StubReLU6(BaseStubReLU6):
  def __init__(self, input=None, output=None, **kwargs):
    super(StubReLU6, self).__init__(input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #     with tf.variable_scope(self.cell_name, 'cell'):
    #       return functools.partial(tf.nn.relu6)(*args, **kwargs)
    # else:
    return functools.partial(tf.nn.relu6)(*args, **kwargs)


class StubSoftmax(BaseStubSoftmax):
  def __init__(self, input=None, output=None, **kwargs):
    super(StubSoftmax, self).__init__(input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #     with tf.variable_scope(self.cell_name, 'cell'):
    #       return functools.partial(tf.nn.softmax)(*args, **kwargs)
    # else:
    return functools.partial(tf.nn.softmax)(*args, **kwargs)


class StubSigmoid(BaseStubSigmoid):
  def __init__(self, input=None, output=None, **kwargs):
    super(StubSigmoid, self).__init__(input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #     with tf.variable_scope(self.cell_name, 'cell'):
    #       return functools.partial(tf.nn.softmax)(*args, **kwargs)
    # else:
    return functools.partial(tf.nn.sigmoid)(*args, **kwargs)


class StubAvgPooling2d(BaseStubAvgPooling2d):
  def __init__(self, kernel_size_h=2, kernel_size_w=2, input=None, output=None, **kwargs):
    super(StubAvgPooling2d, self).__init__(kernel_size_h, kernel_size_w, input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #     with tf.variable_scope(self.cell_name, 'cell'):
    #       return functools.partial(slim.avg_pool2d,
    #                                kernel_size=(self.kernel_size_h,self.kernel_size_w),
    #                                stride=(self.kernel_size_h,self.kernel_size_w))(*args, **kwargs)
    # else:
    return functools.partial(slim.avg_pool2d,
                             kernel_size=(self.kernel_size_h, self.kernel_size_w),
                             stride=(self.kernel_size_h, self.kernel_size_w))(*args, **kwargs)


class StubMaxPooling2d(BaseStubMaxPooling2d):
  def __init__(self, kernel_size_h=2, kernel_size_w=2, input=None, output=None, **kwargs):
    super(StubMaxPooling2d, self).__init__(kernel_size_h, kernel_size_w, input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #     with tf.variable_scope(self.cell_name, 'cell'):
    #       return functools.partial(slim.max_pool2d,
    #                                kernel_size=(self.kernel_size_h,self.kernel_size_w),
    #                                stride=(self.kernel_size_h,self.kernel_size_w))(*args, **kwargs)
    # else:
    return functools.partial(slim.max_pool2d,
                             kernel_size=(self.kernel_size_h, self.kernel_size_w),
                             stride=(self.kernel_size_h, self.kernel_size_w))(*args, **kwargs)


class StubGlobalPooling2d(BaseStubGlobalPooling2d):
  def __init__(self, input=None, output=None, **kwargs):
    super(StubGlobalPooling2d, self).__init__(input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #     with tf.variable_scope(self.cell_name, 'cell'):
    #       return functools.partial(tf.reduce_mean, axis=[1,2])(*args, **kwargs)
    # else:
    return functools.partial(tf.reduce_mean, axis=[1, 2])(*args, **kwargs)


class StubDropout2d(BaseStubDropout2d):
  def __init__(self, rate, input=None, output=None, **kwargs):
    super(StubDropout2d, self).__init__(rate, input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #     with tf.variable_scope(self.cell_name, 'cell'):
    #       return functools.partial(tf.nn.dropout, keep_prob=self.rate)(*args,**kwargs)
    # else:
    return functools.partial(tf.nn.dropout, keep_prob=self.rate)(*args, **kwargs)


class StubInput(BaseStubInput):
  def __init__(self, shape, input=None, output=None, **kwargs):
    super(StubInput, self).__init__(shape, input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    return functools.partial(tf.placeholder, shape=self.shape, dtype=tf.float32)(*args, **kwargs)


class StubBatchNormalization2d(BaseStubBatchNormalization2d):
  def __init__(self, input=None, output=None, **kwargs):
    super(StubBatchNormalization2d, self).__init__(input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #     with tf.variable_scope(self.cell_name, 'cell'):
    #       return functools.partial(slim.batch_norm)(*args, **kwargs)
    # else:
    return functools.partial(slim.batch_norm)(*args, **kwargs)


class StubBilinearResize(BaseStubBilinearResize):
  def __init__(self, height, width, input=None, output=None, **kwargs):
    super(StubBilinearResize, self).__init__(height, width, input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    # if self.block_name != '' and self.cell_name != '':
    #   with tf.variable_scope(self.block_name, 'block'):
    #     with tf.variable_scope(self.cell_name, 'cell'):
    #       return functools.partial(tf.image.resize_bilinear, size=(self.height,self.width))(*args, **kwargs)
    # else:
    return functools.partial(tf.image.resize_bilinear, size=(self.height, self.width))(*args, **kwargs)


class StubIdentity(BaseStubIdentity):
  def __init__(self, input=None, output=None, **kwargs):
    super(StubIdentity, self).__init__(input, output, **kwargs)

  def __call__(self, *args, **kwargs):
    return tf.identity(*args, **kwargs)


class LayerFactory(object):
  def __init__(self):
    pass

  def __getattr__(self, item):
    if item not in ['dense',
                    'conv2d',
                    'separable_conv2d',
                    'concat',
                    'add',
                    'dot',
                    'avg_pool2d',
                    'max_pool2d',
                    'global_pool2d',
                    'relu',
                    'flatten',
                    'relu6',
                    'bn2d',
                    'softmax',
                    'sigmoid',
                    'dropout_2d',
                    'bilinear_resize',
                    'spp',
                    'identity',
                    'input']:
      if item.endswith('_branch'):
        return None

      return getattr(super(LayerFactory, self), item)

    def func(*args, **kwargs):
      if item == 'dense':
        return StubDense(*args, **kwargs)
      elif item == 'conv2d':
        return StubConv2d(*args, **kwargs)
      elif item == 'separable_conv2d':
        return StubSeparableConv2d(*args, **kwargs)
      elif item == 'spp':
        return StubSPP(*args, **kwargs)
      elif item == 'concat':
        return StubConcatenate(*args, **kwargs)
      elif item == 'add':
        return StubAdd(*args, **kwargs)
      elif item == 'dot':
        return StubDot(*args, **kwargs)
      elif item == 'avg_pool2d':
        return StubAvgPooling2d(*args, **kwargs)
      elif item == 'max_pool2d':
        return StubMaxPooling2d(*args, **kwargs)
      elif item == 'global_pool2d':
        return StubGlobalPooling2d(*args, **kwargs)
      elif item == 'relu':
        return StubReLU(*args, **kwargs)
      elif item == 'relu6':
        return StubReLU6(*args, **kwargs)
      elif item == 'flatten':
        return StubFlatten(*args, **kwargs)
      elif item == 'bn2d':
        return StubBatchNormalization2d(*args, **kwargs)
      elif item == 'softmax':
        return StubSoftmax(*args, **kwargs)
      elif item == 'sigmoid':
        return StubSigmoid(*args, **kwargs)
      elif item == 'dropout_2d':
        return StubDropout2d(*args, **kwargs)
      elif item == 'bilinear_resize':
        return StubBilinearResize(*args, **kwargs)
      elif item == 'identity':
        return StubIdentity(*args, **kwargs)
      elif item == 'input':
        return StubInput(*args, **kwargs)

    return func


if __name__ == '__main__':
  layer_factory = LayerFactory()
  input = layer_factory.input(shape=[1,224,224,3], dtype=tf.float32)
  conv2d_1 = layer_factory.conv2d(3, 16, kernel_size=3)
  conv2d_2 = layer_factory.conv2d(16, 32, kernel_size=3)
  conv2d_3 = layer_factory.conv2d(32, 64, kernel_size=3)
  relu = layer_factory.relu()
  conv2d_4 = layer_factory.conv2d(64,2, kernel_size=3)
  flatten = layer_factory.flatten()
  softmax = layer_factory.softmax()

  tf_input = input()
  tf_conv2d_1 = conv2d_1(tf_input)
  tf_conv2d_2 = conv2d_2(tf_conv2d_1)
  tf_conv2d_3 = conv2d_3(tf_conv2d_2)
  tf_relu = relu(tf_conv2d_3)
  tf_conv2d_4 = conv2d_4(tf_relu)
  tf_flatten = flatten(tf_conv2d_4)
  tf_softmax = softmax(tf_flatten)

  with tf.Session() as sess:
    sess.run(tf.global_variables_initializer())
    aa = sess.run(tf_softmax, feed_dict={tf_input: np.random.random([1,224,224,3])})
    print(aa)
  pass





from __future__ import unicode_literals
from __future__ import division

import numpy as np
import unittest

def confusion_matrix(actual,predicated_score):
  '''
  compute confusion matrix (for muti-classification task and binary-classification task)
  Parameters
  ----------
  actual: list or ndarray
          ground truth label list
  predicted_score: list or ndarray
          predicted score (for every label)

  Returns
  -------
  return confusion matrix
  '''
  # type consistent
  if type(actual) == list:
    actual = np.array(actual, dtype=np.int32)
  if type(predicated_score) == list:
    predicated_score = np.array(predicated_score)

  if actual.dtype != np.int32:
    actual = actual.astype(dtype=np.int32)

  # shape consistent
  assert (len(actual.shape) == 1)
  assert (len(predicated_score.shape) == 2)
  assert (actual.shape[0] == predicated_score.shape[0])

  # label number
  class_num = predicated_score.shape[1]

  # confusion matrix
  predicated = np.argmax(predicated_score, axis=1)
  cm = np.zeros((class_num,class_num))
  num = len(actual)
  for i in range(num):
    cm[actual[i],predicated[i]] += 1

  return cm

def compute_confusion_matrix(actual, predict_label, class_num):
  if type(actual) == list:
    actual = np.array(actual, dtype=np.int32)

  if type(predict_label) == list:
    predict_label = np.array(predict_label)

  cm = np.zeros((class_num, class_num))
  for c_i in range(class_num):
    c_i_pos = np.where(actual == c_i)
    c_i_predict = predict_label[c_i_pos]
    for to_i in range(class_num):
      to_i_num = len(np.where(c_i_predict == to_i)[0])
      cm[c_i, to_i] = to_i_num

  return cm

#unit test
class TestConfusionMatrix(unittest.TestCase):
  def test_cm(self):
    actual = [0,0,1,2,1,2]
    predicated_score = [[0.8,0.1,0.1],[0.3,0.5,0.2],[0.1,0.5,0.4],[0.2,0.2,0.6],[0.2,0.3,0.5],[0.2,0.1,0.7]]
    cm = confusion_matrix(actual,predicated_score)
    np.testing.assert_array_almost_equal(cm.flatten(),np.array([1.0,1.0,0.0,0.0,1.0,1.0,0.0,0.0,2.0]))

if __name__ == "__main__":
    unittest.main()


# -*- coding: UTF-8 -*-
# @Time    : 2020/10/26 10:26 上午
# @File    : spider_dataset.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from uuid import uuid4
from antgo.dataflow.dataset.dataset import Dataset
from antgo.ant.utils import *
from antgo.utils import logger
import os
import sys
import cv2
import numpy as nps
import threading
import requests
import re
import uuid
try:
    import queue
except:
    import Queue as queue


class SpiderDataset(Dataset):
  def __init__(self, command_queue=None, dir=None, params=None):
    super(SpiderDataset, self).__init__('test', dir, params)
    self.command_queue = command_queue
    self.waiting_process_queue = queue.Queue()
    self.count = 0
    self.keywords = {}

  def __baidu_download(self, keyward):
    headers = {
        'Accept-Language': 'zh-CN,zh;q=0.8,zh-TW;q=0.7,zh-HK;q=0.5,en-US;q=0.3,en;q=0.2',
        'Connection': 'keep-alive',
        'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:60.0) Gecko/20100101 Firefox/60.0',
        'Upgrade-Insecure-Requests': '1'
    }
    A = requests.Session()
    A.headers = headers
    url = 'https://image.baidu.com/search/flip?tn=baiduimage&ie=utf-8&word=' + keyward + '&pn='

    def __baidu_find_and_download(waiting_process_queue, search_url, session, dir, max_page_num=50):
        t = 0
        num = 0
        while t < max_page_num:
            Url = search_url + str(t)
            t = t+1
            try:
                Result = session.get(Url, timeout=7, allow_redirects=False)
            except BaseException:
                t = t + 60
                continue
            else:
                pic_url = re.findall('"objURL":"(.*?)",', Result.text, re.S)  # 先利用正则表达式找到图片url
                for each in pic_url:
                  print('正在下载第' + str(num + 1) + '张图片，图片地址:' + str(each))
                  try:
                      if each is not None:
                          pic = requests.get(each, timeout=7)
                      else:
                          continue
                  except BaseException:
                      print('错误，当前图片无法下载')
                      continue
                  else:
                      # 分配唯一文件标识
                      file_folder = os.path.join(dir, 'test')
                      if not os.path.exists(file_folder):
                        os.makedirs(file_folder)
                      
                      file_path = os.path.join(file_folder, '%s.jpg'%str(uuid.uuid4()))
                      with open(file_path, 'wb') as fp:
                        fp.write(pic.content)
                      num += 1

                      # 加入等待处理队列
                      waiting_process_queue.put(file_path)

        # 添加结束标记
        waiting_process_queue.put(None)
    
    # 搜索和下载
    t = threading.Thread(target=__baidu_find_and_download, args=(self.waiting_process_queue, url, A, self.dir))
    t.start()

  def __bing_download(self, keyward):
    headers = {
        'Accept-Language': 'zh-CN,zh;q=0.8,zh-TW;q=0.7,zh-HK;q=0.5,en-US;q=0.3,en;q=0.2',
        'Connection': 'keep-alive',
        'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:60.0) Gecko/20100101 Firefox/60.0',
        'Upgrade-Insecure-Requests': '1'
    }
    A = requests.Session()
    A.headers = headers
    #url = 'https://image.baidu.com/search/flip?tn=baiduimage&ie=utf-8&word=' + keyward + '&pn='
    url = 'http://cn.bing.com/images/async?q='+keyward+'&first={0}&count=35&relp=35&lostate=r&mmasync=1&dgState=x*175_y*848_h*199_c*1_i*106_r*0'
    
    def __bing_find_and_download(waiting_process_queue, search_url, session, dir, max_page_num=50):
        t = 0
        num = 0
        while t < max_page_num:
            Url = search_url.format(t*35+1)

            t = t+1
            try:
                Result = session.get(Url, timeout=7, allow_redirects=False)
            except BaseException:
                t = t + 60
                continue
            else:
                pic_url = re.findall('src="(.*?)"', Result.text, re.S)  # 先利用正则表达式找到图片url
                for each in pic_url:
                  print('正在下载第' + str(num + 1) + '张图片，图片地址:' + str(each))
                  try:
                      if each is not None:
                          pic = requests.get(each, timeout=7)
                      else:
                          continue
                  except BaseException:
                      print('错误，当前图片无法下载')
                      continue
                  else:
                      # 分配唯一文件标识
                      file_folder = os.path.join(dir, 'test')
                      if not os.path.exists(file_folder):
                        os.makedirs(file_folder)
                      
                      file_path = os.path.join(file_folder, '%s.jpg'%str(uuid.uuid4()))
                      with open(file_path, 'wb') as fp:
                        fp.write(pic.content)
                      num += 1

                      # 加入等待处理队列
                      waiting_process_queue.put(file_path)
        
        # 结束标记
        waiting_process_queue.put(None)

    # 搜索和下载
    t = threading.Thread(target=__bing_find_and_download, args=(self.waiting_process_queue, url, A, self.dir))
    t.start()

  def __google_download(self, keyward):
    pass

  def __spider_data_source(self, config):
    datasource_address = config['datasource_address']
    datasource_keywards = config['datasource_keywards']

    if datasource_address not in self.keywords:
      self.keywords[datasource_address] = []
    
    if datasource_keywards in self.keywords[datasource_address]:
      logger.error("Duplicate keywords.")
      return

    if datasource_address == 'baidu':
      self.__baidu_download(datasource_keywards)
    elif datasource_address == 'bing':
      self.__bing_download(datasource_keywards)
    elif datasource_address == 'google':
      self.__google_download(datasource_keywards)

  def data_pool(self):
    self.count = 0
    while True:
      try:
        # 1.step 接受爬虫目标指令
        data_pack = self.command_queue.get()
        _, config = data_pack
        self.__spider_data_source(config)

        # 2.step 读取等待处理文件
        image_file = self.waiting_process_queue.get()
        if image_file is None:
          continue

        while True:
          try:
            # read image
            image = cv2.imread(image_file)
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            if image is None:
              logger.error("Fail to parse %s"%image_file)
              # get next image file
              image_file = self.waiting_process_queue.get()
              if image_file is None:
                break

              continue
          except:
            logger.error("Fail to parse %s"%image_file)
            # get next image file
            image_file = self.waiting_process_queue.get()
            if image_file is None:
              break

            continue
          
          # increment 1
          self.count += 1

          # return data
          yield image, {}

          # get next image file
          image_file = self.waiting_process_queue.get()
          if image_file is None:
            break
      except:
        logger.error('Fail receive data info.')

  def finish_process_num(self):
    return self.count

  def waiting_process_num(self):
    return self.waiting_process_queue.qsize()

  def at(self, id):
    raise NotImplemented
  
  def split(self, split_params={}, split_method='holdout'):
    raise NotImplemented

  @property
  def size(self):
    return 10000000



# -*- coding: UTF-8 -*-
# @Time    : 18-9-22
# @File    : batch.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from antgo.dataflow.dataset.spider_dataset import SpiderDataset
from antgo.ant.base import *
from antgo.ant.base import _pick_idle_port
from antgo.task.task import *
from antgo.dataflow.common import *
from antgo.dataflow.basic import *
from antgo.dataflow.recorder import *
from antgo.crowdsource.batch_server import *
from antvis.client.httprpc import *
import json


class AntBatch(AntBase):
    def __init__(self,
                 ant_context,
                 ant_name,
                 ant_host_ip,
                 ant_host_port,
                 token,
                 data_factory,
                 ant_dump_dir,
                 ant_task_config, **kwargs):
        super(AntBatch, self).__init__(ant_name, ant_context, token)

        self.ant_data_source = data_factory
        self.ant_dump_dir = ant_dump_dir
        self.ant_context.ant = self
        self.unlabel = kwargs.get('unlabel', False)
        self.ant_task_config = ant_task_config
        self.context.devices = [int(d) for d in kwargs.get('devices', '').split(',') if d != '']
        self.restore_experiment = kwargs.get('restore_experiment', None)
        self.host_ip = ant_host_ip
        self.host_port = ant_host_port
        self.rpc = None

    def ping_until_ok(self):
        while True:
            content = self.rpc.ping.get()
            if content['status'] != 'ERROR':
                break
            # 暂停5秒钟，再进行尝试
            time.sleep(5)

    def start(self):
        # 1.step 加载挑战任务
        running_ant_task = None
        if self.token is not None:
            # 1.1.step 从平台获取挑战任务配置信息
            response = mlogger.getEnv().dashboard.challenge.get(command=type(self).__name__)
            if response['status'] == 'ERROR':
                logger.error('Couldnt load challenge task.')
                self.token = None
            elif response['status'] == 'SUSPEND':
                # prohibit submit challenge task frequently
                # submit only one in one week
                logger.error('Prohibit submit challenge task frequently.')
                exit(-1)
            elif response['status'] == 'OK':
                content = response['content']

                if 'task' in content:
                    challenge_task = create_task_from_json(content)
                    if challenge_task is None:
                        logger.error('Couldnt load challenge task.')
                        exit(-1)
                    running_ant_task = challenge_task
            else:
                # unknow error
                logger.error('Unknow error.')
                exit(-1)

        if running_ant_task is None:
            # 1.2.step 加载自定义任务配置信息
            custom_task = create_task_from_xml(self.ant_task_config, self.context)
            if custom_task is None:
                logger.error('Couldnt load custom task.')
                exit(-1)
            running_ant_task = custom_task

        assert (running_ant_task is not None)

        # 2.step 获得实验ID
        experiment_uuid = self.context.experiment_uuid
        if self.restore_experiment is not None:
            experiment_uuid = self.restore_experiment

        # 3.step make experiment folder
        logger.info('Build experiment folder.')
        if not os.path.exists(os.path.join(self.ant_dump_dir, experiment_uuid)):
            os.makedirs(os.path.join(self.ant_dump_dir, experiment_uuid))
        experiment_dump_dir = os.path.join(self.ant_dump_dir, experiment_uuid)

        # 4.step 配置web服务基本信息
        # 选择端口
        self.host_port = _pick_idle_port(self.host_port)

        # 配置调用
        self.rpc = HttpRpc("v1", "batch-api", "127.0.0.1", self.host_port)

        # 准备素材资源
        static_folder = '/'.join(os.path.dirname(__file__).split('/')[0:-1])
        batch_static_dir = os.path.join(experiment_dump_dir, 'batch')
        if not os.path.exists(batch_static_dir):
            shutil.copytree(os.path.join(static_folder, 'resource', 'batch'), batch_static_dir)

        # 命令队列
        command_queue = queue.Queue()

        # # 临时测试假数据
        # command_queue.put(('baidu', {'datasource_address':'baidu', 'datasource_keywards': '人像'}))
        
        # 数据标签
        batch_params = getattr(self.context.params, 'predict', None)
        tags = []
        if batch_params is not None:
            tags = batch_params.get('tags', [])

        # 5.step 配置运行
        def _run_batch_process():
            # 等待web服务启动
            self.ping_until_ok()

            # 是否来自已有实验
            if self.restore_experiment is not None and \
                os.path.exists(os.path.join(experiment_dump_dir, '%s.json' % self.restore_experiment)):
                # 加载保存的执行信息
                running_info = {}
                with open(os.path.join(experiment_dump_dir, '%s.json' % self.restore_experiment), 'r') as fp:
                    running_info = json.load(fp)

                # 更新web页面
                self.rpc.config.post(config_data=json.dumps(running_info))
                return

            # 4.step load dataset
            logger.info('Load task dataset and split.')

            dataset_name = ''
            selected_dataset_stages = []
            if len(running_ant_task.dataset_name.split('/')) == 2:
                dataset_name, dataset_stage = running_ant_task.dataset_name.split('/')
                selected_dataset_stages.append(dataset_stage)
            else:
                dataset_name = running_ant_task.dataset_name
                selected_dataset_stages = ['test']

            # 5.step prepare ablation blocks
            logger.info('Prepare model ablation blocks.')
            ablation_blocks = getattr(self.ant_context.params, 'ablation', [])
            if ablation_blocks is None:
                ablation_blocks = []
            for b in ablation_blocks:
                self.ant_context.deactivate_block(b)

            # 6.step infer
            logger.info('Running inference process.')
            for dataset_stage in selected_dataset_stages:
                logger.info('Process dataset: %s.' % dataset_stage)
                # 获得数据集对象
                ant_test_dataset = None
                if dataset_name.startswith('spider'):
                    # 使用SpiderDataset
                    ant_test_dataset = \
                        SpiderDataset(command_queue,
                                        os.path.join(self.ant_data_source, dataset_name), None)
                else:
                    ant_test_dataset = \
                        running_ant_task.dataset(dataset_stage,
                                                    os.path.join(self.ant_data_source, dataset_name),
                                                    running_ant_task.dataset_params)
                # using unlabel
                if self.unlabel:
                    ant_test_dataset = UnlabeledDataset(ant_test_dataset)

                output_dir = os.path.join(experiment_dump_dir, 'batch', 'static', 'data', dataset_stage)
                if not os.path.exists(output_dir):
                    os.makedirs(output_dir)

                # 设置记录器
                def _callback_func(data):
                    record_content = {
                        'experiment_uuid': experiment_uuid, 
                        'dataset': {
                            dataset_stage: [{'data':single_d, 'tag':[]} for single_d in data]
                        },
                        'tags': tags
                    }
                    for data_group in record_content['dataset'][dataset_stage]:
                        for item in data_group['data']:
                            if item['type'] in ['IMAGE','VIDEO','FILE']:
                                item['data'] = 'static/data/%s/record/%s' % (dataset_stage, item['data'])

                    # 添加当前进度
                    record_content['waiting'] = ant_test_dataset.waiting_process_num()
                    record_content['finished'] = ant_test_dataset.finish_process_num()
                    self.rpc.config.post(config_data=json.dumps(record_content))
                    
                    # 保存执行信息
                    if not os.path.exists(os.path.join(experiment_dump_dir, '%s.json' % experiment_uuid)):
                        with open(os.path.join(experiment_dump_dir, '%s.json' % experiment_uuid), 'w') as fp:
                            json.dump(record_content, fp)
                    else:
                        history_running_info = {}
                        with open(os.path.join(experiment_dump_dir, '%s.json' % experiment_uuid), 'r') as fp:
                            history_running_info = json.load(fp)
                        history_running_info['dataset'][dataset_stage].extend(record_content['dataset'][dataset_stage])

                        with open(os.path.join(experiment_dump_dir, '%s.json' % experiment_uuid), 'w') as fp:
                            json.dump(history_running_info, fp)

                self.context.recorder =  LocalRecorderNodeV2(_callback_func)
                with safe_recorder_manager(self.context.recorder):
                    # 完成推断过程
                    self.context.call_infer_process(ant_test_dataset, dump_dir=output_dir)

            return

        process = threading.Thread(target=_run_batch_process)
        process.daemon = True
        process.start()

        # 5.step 启动bach server
        batch_server_start(experiment_dump_dir, self.host_port, command_queue)



# --------------------------------------------------------
# Fast R-CNN
# Copyright (c) 2015 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Written by Ross Girshick
# --------------------------------------------------------

import numpy as np
cimport numpy as np

cdef inline np.float32_t max(np.float32_t a, np.float32_t b):
    return a if a >= b else b

cdef inline np.float32_t min(np.float32_t a, np.float32_t b):
    return a if a <= b else b

def nms(np.ndarray[np.float32_t, ndim=2] dets, np.float thresh):
    cdef np.ndarray[np.float32_t, ndim=1] x1 = dets[:, 0]
    cdef np.ndarray[np.float32_t, ndim=1] y1 = dets[:, 1]
    cdef np.ndarray[np.float32_t, ndim=1] x2 = dets[:, 2]
    cdef np.ndarray[np.float32_t, ndim=1] y2 = dets[:, 3]
    cdef np.ndarray[np.float32_t, ndim=1] scores = dets[:, 4]

    cdef np.ndarray[np.float32_t, ndim=1] areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    cdef np.ndarray[np.int_t, ndim=1] order = scores.argsort()[::-1]

    cdef int ndets = dets.shape[0]
    cdef np.ndarray[np.int_t, ndim=1] suppressed = \
            np.zeros((ndets), dtype=np.int)

    # nominal indices
    cdef int _i, _j
    # sorted indices
    cdef int i, j
    # temp variables for box i's (the box currently under consideration)
    cdef np.float32_t ix1, iy1, ix2, iy2, iarea
    # variables for computing overlap with box j (lower scoring box)
    cdef np.float32_t xx1, yy1, xx2, yy2
    cdef np.float32_t w, h
    cdef np.float32_t inter, ovr

    keep = []
    for _i in range(ndets):
        i = order[_i]
        if suppressed[i] == 1:
            continue
        keep.append(i)
        ix1 = x1[i]
        iy1 = y1[i]
        ix2 = x2[i]
        iy2 = y2[i]
        iarea = areas[i]
        for _j in range(_i + 1, ndets):
            j = order[_j]
            if suppressed[j] == 1:
                continue
            xx1 = max(ix1, x1[j])
            yy1 = max(iy1, y1[j])
            xx2 = min(ix2, x2[j])
            yy2 = min(iy2, y2[j])
            w = max(0.0, xx2 - xx1 + 1)
            h = max(0.0, yy2 - yy1 + 1)
            inter = w * h
            ovr = inter / (iarea + areas[j] - inter)
            if ovr >= thresh:
                suppressed[j] = 1

    return keep



# -*- coding: UTF-8 -*-
# @Time : 2018/4/28
# @File : queue_dataset.py
# @Author: Jian <jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from antgo.dataflow.dataset.dataset import Dataset
import os
import sys
from antgo.utils.serialize import *
from PIL import Image
import imageio
import numpy as np
from io import BytesIO

class QueueDataset(Dataset):
  def __init__(self, queue=None):
    super(QueueDataset, self).__init__('test', '', None)
    self.queue = queue
    self.annotation = {}

  @property
  def size(self):
    return sys.maxsize

  def at(self, id):
    raise NotImplementedError

  def now(self):
    return self.annotation

  def _parse_data(self, data, data_type, request_param):
    assert(data_type not in ['VIDEO'])
    try:
      if data_type == 'IMAGE':
        image_data = Image.open(data)
        img_data = np.fromstring(image_data.tobytes(), dtype=np.uint8)
        img_data = img_data.reshape((image_data.size[1], image_data.size[0], len(image_data.getbands())))
        return (img_data,{}) if request_param is None else (img_data, request_param)
      if data_type == 'IMAGE_MEMORY':
        image_data = Image.open(BytesIO(data))
        img_data = np.fromstring(image_data.tobytes(), dtype=np.uint8)
        img_data = img_data.reshape((image_data.size[1], image_data.size[0], len(image_data.getbands())))
        return (img_data,{}) if request_param is None else (img_data, request_param)
      elif data_type == 'FILE':
        with open(data, 'r') as fp:
          return (fp.read(),{}) if request_param is None else (fp.read(), request_param)
      elif data_type == 'STRING':
        return (data,{}) if request_param is None else (data, request_param)
      elif data_type == 'JSON':
        return (data,{}) if request_param is None else (data, request_param)
      else:
        return (None,{})
    except:
      return None

  def data_pool(self):
    while True:
      data_pack = self.queue.get()
      if type(data_pack) == list and \
              (type(data_pack[0]) == list or type(data_pack[0]) == tuple):
        # multi-data
        for single_pack in data_pack:
          data, data_type, request_param = single_pack
          response_data = self._parse_data(data, data_type, request_param)
          if response_data is None:
            # 解析数据有误，不进行处理
            continue

          # annotation
          self.annotation = response_data[1]
          yield response_data
      else:
        # single-data
        data, data_type, request_param = data_pack
        if data_type == 'VIDEO':
          # 解析视频文件
          try:
            reader = imageio.get_reader(data)
            for im in reader:
              # 读取一帧
              img_data = np.fromstring(im.tobytes(), dtype=np.uint8)
              img_data = img_data.reshape((im.shape[0], im.shape[1], im.shape[2]))
              # 当前annotation
              self.annotation = request_param
              yield (img_data, {}) if request_param is None else (img_data, request_param)
          except:
            # 解析视频有误，不进行处理
            continue
        else:
          response_data = self._parse_data(data, data_type, request_param)
          if response_data is None:
            # 解析数据有误，不进行处理
            continue

          # annotation
          self.annotation = response_data[1]
          yield response_data


# -*- coding: UTF-8 -*-
# @Time    : 17-12-29
# @File    : pck.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import numpy as np
from antgo.task.task import *
from antgo.measures.base import *


class AntPCK(AntMeasure):
  def __init__(self, task):
    super(AntPCK, self).__init__(task, 'PCK')
    assert(task.task_type == 'LANDMARK')
    
    self.joints = self.task.joints
    self.joints_normalizer = self.task.joints_normalizer
    self.is_support_rank = True

  def pcki(self, joint_id, gtJ, prJ, idlh=3, idrs=12):
    """ Compute PCK accuracy on a given joint
    Args:
      joint_id	: Index of the joint considered
      gtJ			: Ground Truth Joint
      prJ			: Predicted Joint
      idlh		: Index of Normalizer (Left Hip on PCK, neck on PCKh)
      idrs		: Index of Normalizer (Right Shoulder on PCK, top head on PCKh)
    Returns:
      (float) NORMALIZED L2 ERROR
    """
    return np.linalg.norm(gtJ[joint_id] - prJ[joint_id][::-1]) / np.linalg.norm(gtJ[idlh] - gtJ[idrs])

  def eva(self, data, label):
    if label is not None:
      data = zip(data, label)
    
    pck_ratio = [[] for _ in range(len(self.joints))]
    pck_flatten_ratio = []
    pck_id = []
    sample_scores = []
    for landmark_predict, gt in data:
      id = gt['id']
      landmark_gt = np.array(gt['landmark'])
      landmark_weight = np.ones((len(self.joints)))
      if len(landmark_gt.shape) == 1:
        landmark_gt = landmark_gt.reshape((-1, 2))
      for i in range(len(self.joints)):
        if landmark_gt[i, 0] < 0 or landmark_gt[i, 1] < 0:
          landmark_weight[i] = 0
      
      landmark_predict = np.array(landmark_predict)
      if len(landmark_predict) == 1:
        landmark_predict = landmark_predict.reshape((-1, 2))
      
      # landmark predict shape (joints, 2)
      # landmark gt shape(joints, 2)
      offset_s = len(pck_flatten_ratio)
      for joint_i in range(len(self.joints)):
        if landmark_weight[joint_i] > 0:
          value = self.pcki(joint_i,
                            landmark_gt,
                            landmark_predict,
                            self.joints_normalizer[0],
                            self.joints_normalizer[1])
          pck_ratio[joint_i].append(value)
          pck_flatten_ratio.append(value)
          pck_id[joint_i].append(id)
      
      offset_e = len(pck_flatten_ratio)
      sample_pck = pck_flatten_ratio[offset_s:offset_e]
      sample_scores.append({'id': id, 'score': float(np.mean(sample_pck))})
    
    pck_total_ratio = float(np.mean(pck_flatten_ratio))
    return {'statistic': {'name': self.name,
                          'value':[{'name': self.name, 'value': pck_total_ratio, 'type': 'SCALAR'}]},
            'info': sample_scores}